/*##############################################################################

nIP - nano IP stack

File        : dispatcher.h

Description : Internetstack State Machine

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _NIP_DISPATCHER_H
#define _NIP_DISPATCHER_H

#include "net/ip.h"
#include "net/net_if.h"
#include "net/tcp.h"

/** Commands to be executed by the dispatcher.
 * @note this list is currently not in use but may replace the func-pointer
 * in the nip_disp_params_t choice at some time in future. */
typedef enum
{
	NIP_DC_NONE,     /** command to indicate empty unused structure */
	NIP_DC_IP_SEND,
	NIP_DC_TCP_SEND, /** */
	NIP_DC_HARD_RECV,
	NIP_DC_IP_RECV,
	NIP_DC_ARP_RECV,
	NIP_DC_ICMP_RECV,
	NIP_DC_TCP_RECV,
	NIP_DC_CUSTOM    /** custom command */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_disp_cmd_t;

/** @name Dispatcher status flags*/
//@{
#define NIP_DISP_RUNNING     0x01 /**< Dispatcher is running. Do not allow another instance. */
#define NIP_DISP_CHECK_IF    0x02 /**< Some network interface flag changed */
#define NIP_DISP_CHECK_UDP   0x04 /**< Check UDP sockets */
#define NIP_DISP_CHECK_TCP   0x08 /**< TCP connections need to be checked. */
#define NIP_DISP_TIMER       0x10 /**< Check stack timer */
#define NIP_DISP_CHECK_TRANS 0x20 /**< check transmissions. same as timer, currently*/
#define NIP_DISP_CHECK_MDNS  0x40 /**< perform mDNS checks */
//@}

/** @name Dispatcher network interface event flags */
//@{
#define NIP_DISP_IF_ETH_R  0x01 /**< Missed Eth Receive Call because Dispatcher was busy. */
// #define NIP_DISP_IF_IP     0x02 /**< IP transmission is waiting for ARP response. */
#define NIP_DISP_IF_TIMER  0x04 /**< check interface timers. */
#define NIP_DISP_IF_DHCP   0x08 /**< check DHCP status for interface, currently same as timer */
#define NIP_DISP_IF_AUTOIP 0x10 /**< check autoIP implementation */
#define NIP_DISP_IF_GROUP  0x20 /**< check host group memberships */
//@}
/** Dispatcher Command Parameters
 * 	@todo switch from function pointers to the above nip_disp_cmd_t and have
 * 	      one global table to be set up in nip_init() which resolves commands
 * 	      to functions. Benefit: Protocolls are easy exchangable, even IP.
 **/
typedef union
{
	struct
	{
		void               (*func)(void); /**< pointer to custom function */
		nip_net_if_id_t    net_if_id;     /**< Network Interface. */
		nip_net_if_trans_t *trans;        /**< pointer to current transmission */
		uint8_t remote_hw_addr[ NIP_MAX_HW_ADDR_SIZE  ]; /**< remote hardware address */
	}
	#ifndef DOXYGEN
	__attribute__((packed))
	#endif
	common; /**< parameter included in all calls. */

	struct
	{
		// --- union head --
		void               (*func)(void); /**< pointer to custom function */
		nip_net_if_id_t    net_if_id;     /**< Network Interface. */
		nip_net_if_trans_t *trans;        /**< pointer to current transmission */
		uint8_t remote_hw_addr[ NIP_MAX_HW_ADDR_SIZE  ]; /**< remote hardware address */
		// ---

		nip_tcp_sock_id_t  sockid;
		uint8_t            flags;
//		uint8_t  local_nw_addr[ NIP_MAX_NET_ADDR_SIZE ]; /**< target network address */
	}
	#ifndef DOXYGEN
	__attribute__((packed))
	#endif
	tcp;

}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_disp_params_t;

/** Information needed to manage the dispatcher state machine. */
typedef struct
{
	nip_disp_params_t next; /**< Next Dispatcher Call */
	uint8_t           flags;/**< Dispatcher flags. */
	uint8_t           if_flags[ NIP_NET_IF_COUNT ]; /** Interface flags */

}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_disp_state_t;

/** @name Macros for easier access to dispatcher variables */
//@{
#define NIP_CURR_CMD nip_disp.next.common.func

#if NIP_NET_IF_COUNT > 1
	#define NIP_CURR_IF     nip_disp.next.common.net_if_id
	#define SET_CURR_IF(id) nip_disp.next.common.net_if_id = id;
#else
	#define NIP_CURR_IF 0
	#define SET_CURR_IF(id)
#endif

//@}

extern nip_disp_state_t nip_disp;

#define nip_disp_notify( events ) nip_disp.flags |= events
void nip_disp_notify_if( nip_net_if_id_t net_if_id, uint8_t event );
// void nip_disp_notify( uint8_t flags );
nip_error_t nip_dispatcher( void );

#endif /* _NIP_DISPATCHER_H */
