/*##############################################################################

nIP - nano IP stack

File        : ip.h

Description : Internet Protocol

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _NIP_IP_H
 #define _NIP_IP_H

#include "nip_init.h"
#include "net_if.h"

/** activate this option to disable memory management for icmp */
// #define NIP_ICMP_TEST_NO_MEM_MGR


#define NIP_IP_DEFAULT_HEADER_SIZE  20
#define NIP_IP_DEFAULT_VERSION      4
#define NIP_IP_TTL  64

#define NIP_IGMP_MAX_REPORT_DELAY  10

#ifndef NIP_IP_MAX_HOST_GROUPS
	#define NIP_IP_MAX_HOST_GROUPS  1
#endif

#if NIP_IP_MAX_HOST_GROUPS <= 255
typedef uint8_t  nip_host_group_id_t;
#elif NIP_IP_MAX_HOST_GRUOPS <= 65536
typedef uint16_t nip_host_group_id_t;
#else
typedef uint32_t nip_host_group_id_t;
#endif


/** \name IP protocol types */
/** see "Assigned Numbers" RFC for complete list */
//@{
#define NIP_IP_PROTO_ICMP    01
#define NIP_IP_PROTO_IGMP    02
#define NIP_IP_PROTO_IP      04
#define NIP_IP_PROTO_TCP     06
#define NIP_IP_PROTO_UDP     17
//@}


typedef uint8_t nip_ip_addr[4];


struct IP_HEADER{
	unsigned int     ihl:4;     // 4-Bit version, 4-Bit header length
	unsigned int     version:4;
	uint8_t     tos;       // type of servive
	uint16_t    total_length;
	uint16_t    id;
	uint16_t    frag_offset;
	uint8_t     ttl;
	uint8_t     protocol;
	uint16_t    checksum;
	nip_ip_addr src_addr;
	nip_ip_addr dst_addr;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

typedef enum
{
	NIP_ICMP_T_ECHO_REP          = 0,
	NIP_ICMP_T_DST_UNREACH       = 3 ,
	NIP_ICMP_T_SRC_QUENCH        = 4,
	NIP_ICMP_T_ECHO_REQ          = 8,
	NIP_ICMP_T_ROUTER_ADV        = 9 ,
	NIP_ICMP_T_ROUTER_SOL        = 10,
	NIP_ICMP_T_TIME_EXCEED       = 11,
	NIP_ICMP_T_PARAM_PROBL       = 12,
	NIP_ICMP_T_TIMESTAMP         = 13,
	NIP_ICMP_T_TIMESTAMP_REP     = 14,
	NIP_ICMP_T_ADDR_MASK_REQ     = 15,
	NIP_ICMP_T_ADDR_MASK_REP     = 18,
	NIP_ICMP_T_TRACEROUTE        = 30,
	NIP_ICMP_T_DATAGRAM_CONV_ERR = 31,
	NIP_ICMP_T_MOBILE_HOST_REDIR = 32,
	NIP_ICMP_T_IPV6_WHERE_ARE_U  = 33,
	NIP_ICMP_T_IPV6_I_AM_HERE    = 34,
	NIP_ICMP_T_MOBILE_REG_REQ    = 35,
	NIP_ICMP_T_MOBILE_REG_RES    = 36,
	NIP_ICMP_T_DOMAIN_NAME_REQ   = 37,
	NIP_ICMP_T_DOMAIN_NAME_REP   = 38,
	NIP_ICMP_T_SKIP              = 39,
	NIP_ICMP_T_PHOTURIS          = 40
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_icmp_type_t;

struct ICMP_HEADER{
	nip_icmp_type_t type;
	uint8_t  code;
	uint16_t checksum;
	uint16_t id;
	uint16_t seq;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

typedef enum
{
	NIP_IP_GROUP_NON = 0,  /** Membership configuration unused */
	NIP_IP_GROUP_DELAYING, /** Notify network about host group membership*/
	NIP_IP_GROUP_IDLE,     /** Group membership in use */
} nip_ip_group_membership_stat_t;


struct nip_ip_group_membership
{

	uint8_t addr[4];         /**< host group address */
	uint8_t member_cnt;      /**< number of members who subscribed to the group*/
	nip_net_if_id_t if_id;   /**< network interface to receive multicasts on */
	#if NIP_IGMP_ENABLE == 1
	nip_time_t      t_report;/**< time at which to report in DELAYING state */
	#endif
	nip_ip_group_membership_stat_t state;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

#define NIP_IGMP_VERTYPE_QUERY  0x11
#define NIP_IGMP_VERTYPE_REPORT 0x12

struct nip_igmp_msg
{
	uint8_t  ver_type;
	uint8_t  unused;
	uint16_t checksum;
	uint8_t  addr[4];
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

void nip_ip_disp_receive( void );
void nip_ip_disp_send( void );

extern uint8_t nip_ip_localhost[4];
extern uint8_t nip_ip_broadcast[4];
extern uint8_t nip_ip_null[4];

void nip_ip_addr_set     ( nip_net_if_id_t if_id, uint8_t *addr );
void nip_ip_subnet_set   ( nip_net_if_id_t if_id, uint8_t *mask );
void nip_ip_addr_autoconf( nip_net_if_id_t if_id );

nip_net_if_trans_t *nip_ip_route( uint8_t *target_addr, nip_net_if_id_t if_id,
	nip_net_if_trans_t *trans );

nip_error_t nip_ip_join_host_group( nip_net_if_id_t if_id, uint8_t *addr );
void nip_ip_disp_check_group( void );
void nip_igmp_disp_receive( void );
void nip_ip_query_groups( nip_net_if_id_t if_id );

void nip_ip_up( nip_net_if_id_t if_id );
void nip_ip_down( nip_net_if_id_t if_id );

#endif /* _NIP_IP_H */
