/*##############################################################################

nIP - nano IP stack

File        : tcp.h

Description : Transport Control Protocol

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _NIP_TCP_H
 #define _NIP_TCP_H

#include "nip_init.h"
#include "mem.h"

#ifndef NIP_TCP_MAX_SOCKETS
	/** Define maximum number of TCP sockets for static socket array. */
	///@todo check why setting this to 1 leads to massive compile errors
	#define NIP_TCP_MAX_SOCKETS      2
#endif

#ifndef NIP_TCP_MAX_CONNECTIONS
	/** Define maximum number of TCP sockets for dynamic connection array. */
	#define NIP_TCP_MAX_CONNECTIONS  5
#endif

#ifndef NIP_TCP_MAX_UNACCEPTED
	/** Define maximum number of unaccepted TCP connections */
	#define NIP_TCP_MAX_UNACCEPTED   5
#endif

#ifndef NIP_TCP_DEFAULT_RCV_WND
	/** Default Receive Window Size. Every IP host should be able to receive
	 * 576 Byte long IP packets. Assuming standard headers, that results in 536
	 * Byte for TCP payload.
	 */
	#define NIP_TCP_DEFAULT_RCV_WND      150
#endif

#ifndef NIP_TCP_MSL
	/** maximum segment lifetime */
	#define NIP_TCP_MSL  120
#endif

#ifndef NIP_TCP_RTO_UBOUND
	/** Upper bound on retransmission timeout (RTO). This value MUST NOT exceed
	 * 255, due to data type restrictions.
	*/
	#define NIP_TCP_RTO_UBOUND  60
#endif

#ifndef NIP_TCP_RTO_LBOUND
	/** Lower bound on retransmission timeout (RTO). This value MUST NOT exceed
	 * 255, due to data type restrictions. */
	#define NIP_TCP_RTO_LBOUND   1
#endif

#ifndef NIP_TCP_RTO_ALPHA
	/** Smoothing factor for round-trip-time. RFC 793 proposes 0.8 or 0.9 */
	#define NIP_TCP_RTO_ALPHA    1
#endif
	#define NIP_TCP_RTO_ALPHA_NEG 0 /* 1-NIP_TCP_RTO_ALPHA */

#ifndef NIP_TCP_RTO_BETA
	/** Delay variance factor for RTO. RFC 793 proposes 1.3 to 2.0 */
	#define NIP_TCP_RTO_BETA     2
#endif


#if   ( NIP_TCP_MAX_CONNECTIONS < 0xFF   )
	typedef uint8_t  nip_tcp_sock_id_t;
	typedef uint8_t  nip_tcp_sock_cnt_t;
	#define NIP_TCP_NO_SOCKET   0xFF
#elif ( NIP_TCP_MAX_CONNECTIONS < 0xFFFF )
	typedef uint16_t nip_tcp_sock_id_t;
	typedef uint16_t nip_tcp_sock_cnt_t;
	#define NIP_TCP_NO_SOCKET   0xFFFF
#else
	typedef uint32_t nip_tcp_sock_id_t;
	typedef uint32_t nip_tcp_sock_cnt_t;
	#define NIP_TCP_NO_SOCKET   0xFFFFFFFF
#endif

#define NIP_TCP_SOCK_FLG_USED      0x01 /**< set to mark socket as used */
#define NIP_TCP_SOCK_FLG_NON_BLOCK 0x02 /**< set to mark non-blocking socket */
#define NIP_TCP_SOCK_FLG_LISTENING 0x04 /**< wait for incoming connections? */
#define NIP_TCP_SOCK_FLG_ACCEPTED  0x08 /**< user accepted connection */
#define NIP_TCP_SOCK_FLG_RECV      0x10 /**< socket ready to receive */
#define NIP_TCP_SOCK_FLG_SEND      0x20 /**< socket ready to send */
#define NIP_TCP_SOCK_FLG_SHUT_WR   0x40 /**< quit writing data on socket */
#define NIP_TCP_SOCK_FLG_SHUT_RD   0x80 /**< quit receiving data on socket */
#define NIP_TCP_SOCK_FLG_AUTOCLOSE (NIP_TCP_SOCK_FLG_SHUT_WR|NIP_TCP_SOCK_FLG_SHUT_RD)

#define NIP_TCP_MIN_HEADER_SIZE    20   /**< minimum size of TCP header */
#define NIP_TCP_MIN_DATA_OFFSET    0x50
#define NIP_TCP_RESET_ACK          0xFFFFFFFF

#define NIP_TCP_SHUT_RD   NIP_TCP_SOCK_FLG_SHUT_RD /**< shut down packet receival */
#define NIP_TCP_SHUT_WR   NIP_TCP_SOCK_FLG_SHUT_WR /**< shut down packet sending */
#define NIP_TCP_SHUT_RDWR (NIP_TCP_SHUT_RD|NIP_TCP_SHUT_WR)

struct nip_tcp_sock_addr
{
	uint16_t port;  /**< TCP port in network byte order */
	uint8_t  ip[4]; /**< IP address */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

struct nip_tcp_sock
{
	nip_tcp_sock_id_t        sockid;    /**< socket id                    */
	uint8_t                  flags;     /**< socket configuration         */
	struct nip_tcp_sock_addr addr;      /**< listening address + port     */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

typedef enum nip_tcp_stat
{
	NIP_TCP_STAT_CLOSED       = 0,
	NIP_TCP_STAT_LISTEN       = 1,
	NIP_TCP_STAT_SYN_RECEIVED = 2,
	NIP_TCP_STAT_SYN_SENT     = 3,
	NIP_TCP_STAT_ESTABLISHED  = 4,
	NIP_TCP_STAT_FIN_WAIT1    = 5,
	NIP_TCP_STAT_FIN_WAIT2    = 6,
	NIP_TCP_STAT_CLOSE_WAIT   = 7,
	NIP_TCP_STAT_CLOSING      = 8,
	NIP_TCP_STAT_LAST_ACK     = 9,
	NIP_TCP_STAT_TIMEWAIT     = 10
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_tcp_stat_t;

typedef uint32_t nip_tcp_seq_t;

/** Transmission Control Block, any TCP connection has to have one
 */
struct nip_tcb
{
	nip_tcp_sock_id_t        connid; /**< ID to address this connection. It may
												 * differ from sockid for connections spawned
												 * from listening socket. Hand this ID to
												 * the application upon accepting new
												 * connections.
												 */
	nip_tcp_sock_id_t        sockid; /**< association to local Socket Array */
	struct nip_tcp_sock_addr rem_addr;
	nip_tcp_stat_t           status;
	uint8_t                  flags;

	nip_mem_handle_t         rcv_buf;
	nip_mem_handle_t         snd_buf;
// 	nip_mem_size_t           rcv_buf_pos;
// 	nip_mem_size_t           snd_buf_pos;

	// Send Sequence Variables
	uint32_t snd_una;
	uint32_t snd_nxt;
	uint16_t snd_wnd;
	uint16_t snd_maxsegsize;
	uint32_t snd_up;
	uint32_t snd_wl1;
	uint32_t snd_wl2;
	uint32_t iss;

	// Receive Sequence Variables
	uint32_t rcv_nxt;
	uint16_t rcv_wnd;
	uint32_t rcv_up;
	uint32_t irs;

	// retransmission handling
	uint8_t    backup;   /**< time to backup before transmitting again */
	uint8_t    srtt;     /**< smoothed round trip time */
	nip_time_t rto;      /**< retransmission timeout */
	uint32_t   rt_una;   /**< longest unacked segment to be retransmitted */
	uint8_t    rt_count; /**< retransmission counter since last ACK */
	uint32_t   push_ptr; /**< remember what sequence number needed to be pushed */
	nip_time_t timeout;  /**< time-wait timeout or user timeout, depending on state */
// 	struct TCP_RETRANSMIT retransmit_stat;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

#define NIP_TCP_FIN_FLAG   0x01
#define NIP_TCP_SYN_FLAG   0x02
#define NIP_TCP_RST_FLAG   0x04
#define NIP_TCP_PSH_FLAG   0x08
#define NIP_TCP_ACK_FLAG   0x10
#define NIP_TCP_URG_FLAG   0x20

// TCP_HEADER
struct nip_tcp_header
{
	uint16_t srcport;
	uint16_t dstport;
	uint32_t seq;
	uint32_t ack;
	uint8_t  dataoffset; // upper 4 bits is count the # of 32Bit words in Header
	uint8_t  flags;
	uint16_t window;
	uint16_t checksum;
	uint16_t urgent;
	uint8_t  options[4]; // upper bytes are Options, lower bytes padding
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

typedef enum
{
	NIP_TCP_CONN_OP_ADD,
	NIP_TCP_CONN_OP_DEL,
	NIP_TCP_CONN_OP_FIND
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_tcp_conn_op_t;

nip_tcp_sock_id_t nip_tcp_socket( struct nip_tcp_sock_addr *addr, uint8_t flags );
void nip_tcp_close( nip_tcp_sock_id_t connid );
void nip_tcp_disp_check( void );
void nip_tcp_disp_receive( void );
nip_tcp_sock_id_t nip_tcp_accept( nip_tcp_sock_id_t sockid );
nip_mem_size_t nip_tcp_read( nip_tcp_sock_id_t connid, uint8_t *buf, nip_mem_size_t size );
nip_mem_size_t nip_tcp_write( nip_tcp_sock_id_t connid, uint8_t *buf, nip_mem_size_t size );
#endif /* _NIP_TCP_H */
