/*##############################################################################

nIP - nano IP stack

File        : udp.h

Description : User Datagram Protocol

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#ifndef _NIP_UDP_H
 #define _NIP_UDP_H

#include "nip_init.h"
#include "mem.h"

#ifndef NIP_UDP_MAX_SOCKETS
	/** Define maximum number of UDP sockets for dynamic socket array. */
	#define NIP_UDP_MAX_SOCKETS        255
#endif

#ifndef NIP_UDP_FIXED_SOCKETS
	/** Fixed number of UDP sockets to be used for fixed socket configuration */
	#define NIP_UDP_FIXED_SOCKETS      4
#endif

#ifndef NIP_UDP_USE_FIXED_SOCKETS
	/** Use fixed socket array (static allocation of memory for fixed numer of
	 * sockets) or use dynamic array of sockets with at most NIP_UDP_MAX_SOCKETS
	 * sockets. */
	#define NIP_UDP_USE_FIXED_SOCKETS  1
#endif

#ifndef NIP_UDP_SOCKET_CHECK_INTVL
	/** Any socket will not be checked for changes twice within the interval,
	 * unless its timer has been reset.*/
	#define NIP_UDP_SOCKET_CHECK_INTVL 1
#endif

#ifndef NIP_UDP_PACKET_QUEUE_SIZE
	/** Each entry in packet queue has cost of sizeof(struct nip_udp_buffer) */
	#define NIP_UDP_PACKET_QUEUE_SIZE  4
#endif

#if   ( NIP_UDP_MAX_SOCKETS <  0xFF  )
	typedef uint8_t  nip_udp_sock_id_t;
	typedef uint8_t  nip_udp_sock_cnt_t;
	#define NIP_UDP_NO_SOCKET   0xFF
#elif ( NIP_UDP_MAX_SOCKETS < 0xFFFF )
	typedef uint16_t nip_udp_sock_id_t;
	typedef uint16_t nip_udp_sock_cnt_t;
	#define NIP_UDP_NO_SOCKET   0xFFFF
#else
	typedef uint32_t nip_udp_sock_id_t;
	typedef uint32_t nip_udp_sock_cnt_t;
	#define NIP_UDP_NO_SOCKET   0xFFFFFFFF
#endif

/** End of queue. To be used to invalidate queue pointer */
#define NIP_UDP_EOQ          NIP_UDP_PACKET_QUEUE_SIZE

#if   ( NIP_UDP_PACKET_QUEUE_SIZE <  255  )
	typedef uint8_t  nip_udp_queue_id_t;
	typedef uint8_t  nip_udp_queue_cnt_t;
#elif ( NIP_UDP_PACKET_QUEUE_SIZE < 65535 )
	typedef uint16_t nip_udp_queue_id_t;
	typedef uint16_t nip_udp_queue_cnt_t;
#else
	typedef uint32_t nip_udp_queue_id_t;
	typedef uint32_t nip_udp_queue_cnt_t;
#endif

#define NIP_UDP_SOCK_FLG_USED      0x01 /**< socket in use? */
#define NIP_UDP_SOCK_FLG_NON_BLOCK 0x02 /**< non-blocking socket? to be set by user */
#define NIP_UDP_SOCK_FLG_LISTENING 0x04 /**< bind socket to port? to be set by user */

typedef enum
{
	NIP_UDP_RX_BUF,
	NIP_UDP_TX_BUF
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
nip_udp_buf_type_t;

struct nip_udp_sock_addr
{
	uint16_t port;
	uint8_t  ip[4];
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

/** Item of UDP Packet Queue. Item is considered unused, when buf equals
 * NIP_MEM_NULL*/
struct nip_udp_buffer
{
	nip_udp_sock_id_t  sockid;
	nip_mem_size_t     size;         /**< current size of packet in buffer */
	nip_mem_handle_t   buf;          /**< dynamic buffer handle */
	nip_udp_queue_id_t next;         /**< next queue item for this socket */
	struct nip_udp_sock_addr addr;  /**< remote address + port        */
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

struct nip_udp_sock
{
	nip_udp_sock_id_t        sockid;    /**< socket id                    */
	uint8_t                  flags;     /**< socket configuration         */
	struct nip_udp_sock_addr addr;      /**< listening address + port     */
	nip_mem_handle_t         tx_packet; /**< handle for transmit buffer   */

	nip_udp_queue_id_t       rx_packet; /**< current item-id in incoming queue*/

}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

struct nip_udp_header{
	uint16_t src_port;
	uint16_t dst_port;
	uint16_t length;
	uint16_t checksum;
}
#ifndef DOXYGEN
__attribute__((packed))
#endif
;

nip_udp_sock_id_t nip_udp_socket( struct nip_udp_sock_addr *addr, uint8_t flags );
void nip_udp_close( nip_udp_sock_id_t sockid );

void nip_udp_disp_send( void );
void nip_udp_disp_receive( void );

nip_error_t nip_udp_init_send( nip_udp_sock_id_t sockid, nip_mem_size_t size );
nip_error_t nip_udp_write( nip_udp_sock_id_t sockid, uint8_t *buf, nip_mem_size_t size);
nip_error_t nip_udp_transmit( nip_udp_sock_id_t sockid, struct nip_udp_sock_addr *dest_addr, nip_net_if_id_t if_id );
nip_mem_size_t nip_udp_recvfrom( nip_udp_sock_id_t sockid, uint8_t *buf, nip_mem_size_t size, struct nip_udp_sock_addr *from_addr );

void nip_udp_forward( nip_udp_sock_id_t sockid );
void nip_udp_cancel_send( nip_udp_sock_id_t sockid );
void nip_udp_cancel_recv( nip_udp_sock_id_t sockid );

nip_mem_handle_t nip_udp_buf_handle( nip_udp_sock_id_t sockid, nip_udp_buf_type_t buftype );
void *nip_udp_data_ptr( nip_udp_sock_id_t sockid, nip_udp_buf_type_t buftype );
void nip_udp_data_mem_ptr( nip_udp_sock_id_t sockid, nip_udp_buf_type_t buftype, nip_mem_ptr_t *res );
void nip_udp_ptr_release( nip_udp_sock_id_t sockid, nip_udp_buf_type_t buftype );

#endif // _NIP_UDP_H
