/*##############################################################################

nIP - nano IP stack

File        : dhcp.c

Description : Dynamic Host Configuration Protocol Client

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#include "nip_init.h"
#include "dispatcher.h"
#include "inet.h"
#include "os_core.h"
#include "nip_error.h"
#include "net/net_if.h"
#include "net/udp.h"
#include "app/dhcp.h"
#include "app/ports.h"


#if (NIP_DHCP_ENABLE==1)

#define DHCPMSGSIZE sizeof( struct nip_dhcp_packet )

/** dhcp configuration variables for each interface */

#if NIP_NET_IF_COUNT > 1
	struct nip_dhcp_config nip_dhcp_conf[ NIP_NET_IF_COUNT ];
	#define NIP_DHCP_CONF( if_id ) nip_dhcp_conf[if_id]
#else
	struct nip_dhcp_config nip_dhcp_conf;
	#define NIP_DHCP_CONF( if_id ) nip_dhcp_conf
#endif
nip_udp_sock_id_t      nip_dhcp_sock = NIP_UDP_NO_SOCKET;


/** enable dhcp for given network interface */
void nip_dhcp_start( nip_net_if_id_t net_if )
{
	NIP_DHCP_CONF( net_if ).state = NIP_DHCP_STAT_INIT;

	nip_disp_notify_if( net_if, NIP_DISP_IF_DHCP );
}

/** disable dhcp for given network interface */
void nip_dhcp_disable( nip_net_if_id_t net_if )
{
	NIP_DHCP_CONF( net_if ).state = NIP_DHCP_STAT_DISABLED;
}

/** Dispatcher function to initialize DHCP UDP socket.
 *
 * On successfull socket setup the function will set the next-pointer to
 * nip_dhcp_disp_check(). The socket setup may fail if the nip stack is out of
 * space for UDP sockets. In that case nip_dhcp_disp_check() will not be called.
 */
void nip_dhcp_disp_sock_init( void )
{
	struct nip_udp_sock_addr addr;

	// Set socket to listen DHCP CLIENT port on all interfaces
	addr.port = NIP_UDP_PORT_DHCPCLIENT;
	addr.ip[0] = 0;
	addr.ip[1] = 0;
	addr.ip[2] = 0;
	addr.ip[3] = 0;

	nip_dhcp_sock = nip_udp_socket(
		&addr,
		NIP_UDP_SOCK_FLG_LISTENING | NIP_UDP_SOCK_FLG_NON_BLOCK
	);

	// if socket was created successfully, continue with nip_dhcp_disp_check()
/*	if ( nip_dhcp_sock != NIP_UDP_NO_SOCKET )
	{
		NIP_CURR_CMD = &nip_dhcp_disp_check;
	}*/
}

/** @return pointer to option in DHCP message or NULL if option was not found. */
struct nip_dhcp_option *nip_dhcp_find_option( uint8_t opttype, struct nip_dhcp_packet *msg, nip_mem_size_t msg_size )
{
	uint8_t  i = 0;


	while ( msg->options[i] != DHCP_OPTION_END )
	{
		// out of space (buffer overflow?)
		if ( &msg->options[i+msg->options[i+1]+2] > &((uint8_t*)msg)[msg_size] )
			break;

		// padding?
		if ( msg->options[i] == DHCP_OPTION_PADDING )
		{
			i++;
			continue;
		}

		// check option at point i
		if (msg->options[i]==opttype){
			// search was successful
			return (struct nip_dhcp_option*) &msg->options[i];
		} else{
			// advance to next option
			i+=msg->options[i+1]+2;
		}
	}

	return NULL;
}

/** Dispatcher function to be called to check dhcp status.
 */
void nip_dhcp_disp_check( void )
{
	nip_net_if_id_t         if_id  = NIP_CURR_IF;
	nip_net_if_t           *net_if = & nip_net_if_list[ if_id ];

	// use different mechanisms to access dhcp configuration for the single and
	// the multi interface cases to minimize use of program memory in either case
#if NIP_NET_IF_COUNT > 1
	struct nip_dhcp_config *conf   = & NIP_DHCP_CONF( if_id );
	#define CONF_ATTR(attr) conf-> attr
#else
	#define CONF_ATTR(attr) nip_dhcp_conf.attr
#endif

	struct nip_dhcp_packet *msg;
	struct nip_dhcp_option *opt;
	nip_mem_size_t         msg_size;
	uint8_t                i;
	uint8_t                optcode;
	void                   *addr;
	nip_time_t             tc = nip_tickcount();

	// ***********************************************************************
	// C h e c k   i n c o m i n g   m e s s a g e s
	while ( 0 < (msg_size = nip_udp_recvfrom( nip_dhcp_sock, NULL, 0, &CONF_ATTR(serv_addr) )) )
	{

		// get pointer to packet data
		msg = nip_udp_data_ptr( nip_dhcp_sock, NIP_UDP_RX_BUF );

		if ( msg != NULL )
		{
			opt = nip_dhcp_find_option( DHCP_OPTION_MSGTYPE, msg, msg_size );

			// check message xid, message op-code and type
			if ( msg->xid == CONF_ATTR(xid)
				&& msg->op == NIP_DHCP_OP_REPLY
				&& opt != NULL )
			{
				switch ( opt->value.msgtype )
				{
					case DHCP_OPTION_MSGTYPE_OFFER:
						if ( CONF_ATTR(state) == NIP_DHCP_STAT_SELECTING )
						{
							// build Request
							opt->value.msgtype = DHCP_OPTION_MSGTYPE_REQUEST;
							nip_memcpy( msg->ciaddr, msg->yiaddr, 4 );
							nip_memset( msg->yiaddr, 0, 4 );

							// forward msg to outgoing queue. The msg pointer stays the
							// same but now stands for the transmission buffer and will
							// be freed later, before transmission.
							nip_udp_forward( nip_dhcp_sock );

							CONF_ATTR(state) = NIP_DHCP_STAT_REQUESTING;
							goto transmit;
						}
						break;
					case DHCP_OPTION_MSGTYPE_ACK:

						// check if current state accepts ACK messages
						if ( CONF_ATTR(state) == NIP_DHCP_STAT_REQUESTING
							|| CONF_ATTR(state) == NIP_DHCP_STAT_REBINDING
							|| CONF_ATTR(state) == NIP_DHCP_STAT_RENEWING )
						{
							// configure network interface
							// set IP address
							nip_ip_addr_set( if_id, msg->yiaddr );
							// set Subnet mask, if that option has been set by server
							opt = nip_dhcp_find_option( DHCP_OPTION_SUBNETMASK, msg, msg_size );
							if ( opt != NULL && opt->valsize == 4  )
								nip_ip_subnet_set( if_id, opt->value.ipaddr );

							// read options for router, nameserver and lease times
							i = 1;
							do
							{
								// save condition for i==1
								addr    = net_if->ip_conf.std_gateway;
								optcode = DHCP_OPTION_ROUTER;

								if ( i==2 )
								{
#if NIP_DNS_ENABLE == 1
									addr    = net_if->ip_conf.name_server;
									optcode = DHCP_OPTION_DNS;
#else
									i++;
#endif
								}
								if ( i==3 )
								{
									addr    = &CONF_ATTR(leasetime);
									optcode = DHCP_OPTION_LEASETIME;
								}
								if ( i==4 )
								{
									addr    = &CONF_ATTR(t1);
									optcode = DHCP_OPTION_RENEWAL_TIME;
								}
								if ( i==5 )
								{
									addr    = &CONF_ATTR(t2);
									optcode = DHCP_OPTION_REBINDING_TIME;
								}
								opt = nip_dhcp_find_option( optcode, msg, msg_size );
								if ( opt != NULL && opt->valsize >= 4 )
								{
									if ( i<3 )
										nip_memcpy( addr, opt->value.ipaddr, 4 );
									else
										*(nip_time_t*)addr = tc + ntohl( opt->value.time );
								} else
								{
									nip_memset( addr, 0, 4 );
								}
							}
							while ( ++i < 6 );

							CONF_ATTR(state) = NIP_DHCP_STAT_BOUND;

						}
						break;

					case DHCP_OPTION_MSGTYPE_NAK:
						// reset to INIT state to discover other server
						CONF_ATTR(state) = NIP_DHCP_STAT_INIT;
						break;
				}
			}

			nip_udp_ptr_release( nip_dhcp_sock, NIP_UDP_RX_BUF );
		}

		// done with message? -> discard
		nip_udp_cancel_recv( nip_dhcp_sock );

		// close UDP socket when in bound state
		if ( CONF_ATTR(state) == NIP_DHCP_STAT_BOUND )
		{
			nip_udp_close( nip_dhcp_sock );
			nip_dhcp_sock = NIP_UDP_NO_SOCKET;
		}

	}


	// ***********************************************************************
	// C h e c k   D H C P   t i m e o u t s
	switch ( CONF_ATTR(state) )
	{
		case NIP_DHCP_STAT_BOUND:

			// check timer T1
			if ( CONF_ATTR(t1) < tc )
			{
				CONF_ATTR(state) = NIP_DHCP_STAT_RENEWING;
				// send request
				goto prepare;
			}
			goto quit;

		case NIP_DHCP_STAT_RENEWING:

			// check timer T2
			if ( CONF_ATTR(t2) < tc )
			{
				CONF_ATTR(state) = NIP_DHCP_STAT_REBINDING;
				// send request
				goto prepare;
			}
			break;

		case NIP_DHCP_STAT_REBINDING:
			// lease expired?
			if ( CONF_ATTR(leasetime) < tc )
			{
				// restart IP address retrieval process
				// DHCP state will be reset within the nnextip_dhcp_start() function
// 				CONF_ATTR(state) = NIP_DHCP_STAT_INIT;
				nip_ip_addr_autoconf( if_id );
				goto prepare;
			}
			break;

		case NIP_DHCP_STAT_REQUESTING:
		case NIP_DHCP_STAT_SELECTING:
			break;

		case NIP_DHCP_STAT_INIT:
			goto prepare;

		default:
				// unhandled state -> quit
			goto quit;
			break;
	}

	// all the above handled states will jump some place else if something
	// special is to happen. -> if we get here, retransmissions have to be checked
	goto retrans;



	// ***********************************************************************
	// M e s s a g e   B u i l d i n g
prepare:

		// restart exponential backoff algorithm
		CONF_ATTR(backoff_delay) = NIP_DHCP_INIT_BACKOFF;
		CONF_ATTR(ts) = 0;

retrans:

		if ( tc < CONF_ATTR(ts) )
		{
			return;
		}

		// Init socket
		if ( nip_dhcp_sock == NIP_UDP_NO_SOCKET )
		{
			NIP_CURR_CMD = &nip_dhcp_disp_sock_init;
			goto quit;
		}

		// initialize send buffer
		if ( NIP_E_OK != nip_udp_init_send( nip_dhcp_sock, DHCPMSGSIZE ) )
			goto quit;

		// allocate uninitialized memory
		if ( NIP_E_OK != nip_udp_write( nip_dhcp_sock, NULL, DHCPMSGSIZE ) )
		{
			nip_udp_cancel_send( nip_dhcp_sock );
			goto quit;
		}

		// Requests aus dem REQUESTING State können nicht neugesendet werden
		// da das Offer nicht gespeichert wird. -> Es wird wieder beim Discover
		// begonnen. Gleichermaßen wird auf das Timeout im SELECTING state
		// reagiert, welcher bedeutet, dass es kein offer gab.
		if ( CONF_ATTR(state) == NIP_DHCP_STAT_REQUESTING
		  || CONF_ATTR(state) == NIP_DHCP_STAT_SELECTING  )
			CONF_ATTR(state) = NIP_DHCP_STAT_INIT;

		if ( CONF_ATTR(state) == NIP_DHCP_STAT_INIT
		  || CONF_ATTR(state) == NIP_DHCP_STAT_REBINDING )
		{
			// Set server address to broadcast and set server port
			nip_memset( CONF_ATTR(serv_addr).ip, 255, 4 );
			CONF_ATTR(serv_addr).port = NIP_UDP_PORT_DHCPSERVER;
		}

		if ( CONF_ATTR(state) == NIP_DHCP_STAT_RENEWING
		  || CONF_ATTR(state) == NIP_DHCP_STAT_REBINDING )
		{
			optcode     = DHCP_OPTION_MSGTYPE_REQUEST;
		}
		else
		//if ( CONF_ATTR(state) == NIP_DHCP_STAT_INIT )
		{
			CONF_ATTR(state) = NIP_DHCP_STAT_SELECTING;
			optcode     = DHCP_OPTION_MSGTYPE_DISCOVER;
			// reset interface's IP address
			// commented, because it seems we don't have to and actually should not
			// reset the IP address, to not interfear with autoIP. Moved this to
			// nip_ip_addr_autoconf() function.
			//nip_memset( net_if->ip_conf.addr, 0, 4 );
		}

		// initialize DHCP Transaction ID with interface id
		/// @todo use some unique number (part of mac address?)
		CONF_ATTR(xid) = net_if->phy_conf.hw_addr[net_if->phy_conf.hw_addr_size-1] + (tc<<8);

		// get pointer to packet data
		msg = nip_udp_data_ptr( nip_dhcp_sock, NIP_UDP_TX_BUF );

		// fill msg with data
		nip_memset( msg, 0, DHCPMSGSIZE );

		msg->htype = net_if->phy_conf.type;
		msg->hlen  = net_if->phy_conf.hw_addr_size;
		msg->xid   = CONF_ATTR(xid);

		// fill in client internet address
		nip_memcpy(
			msg->ciaddr,
			net_if->ip_conf.addr,
			4
		);
		msg->magic_cookie = NIP_DHCP_MAGIC_COOKIE;

		// Set Options
		msg->options[0] = DHCP_OPTION_MSGTYPE;
		msg->options[1] = 1;
		msg->options[2] = optcode;
		msg->options[3] = DHCP_OPTION_END;



transmit:
		// make sure the msg-pointer has been obtained before jumping here

		// set values unique to all client messages
		msg->op    = NIP_DHCP_OP_REQUEST;

		// fill in client hardware address
		nip_memcpy(
			msg->chaddr,
			net_if->phy_conf.hw_addr,
			net_if->phy_conf.hw_addr_size
		);

		// set broadcast flag, because our IP implementation won't accept
		// non-matching unicasts
		msg->flags[0]=0xF0;

		// release pointer to packet data
		nip_udp_ptr_release( nip_dhcp_sock, NIP_UDP_TX_BUF );

		// send DHCP message over given network interface to configured server addr
		nip_udp_transmit( nip_dhcp_sock, &CONF_ATTR(serv_addr), if_id );

		// double backoff delay
		CONF_ATTR(ts) = tc + CONF_ATTR(backoff_delay);
		CONF_ATTR(backoff_delay) <<= 1;
		if ( CONF_ATTR(backoff_delay) > NIP_DHCP_MAX_BACKOFF )
			CONF_ATTR(backoff_delay) = NIP_DHCP_MAX_BACKOFF;

quit:
		return;

}






#endif // (NIP_DHCP_ENABLE==1)
