/*##############################################################################

nIP - nano IP stack

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/
/** @file dns_sd.c
 *
 * DNS Services Discovery Extension.
 *
 * Provides means to register and unregister services.
 *
 */

#include "nip_init.h"
#include "app/dns_sd.h"
#include "app/mdns.h"
#include "mem.h"
#include "nip_error.h"

#if ( NIP_DNSSD_ENABLE == 1 )

extern nip_mdns_namestring_id_t nip_mdns_domain;
extern nip_mdns_cache_id_t      nip_mdns_hostname;
extern nip_mem_ptr_t            nip_mdns_cache;

// Current Service. This should be a list, but to get started, we'll just
// register one service per device.
struct nip_dnssd_service nip_dnssd_services[NIP_DNSSD_MAX_SERVICES];

/** Register Service for the local host on the .local domain
 * @param instance  Instance label e.g. "#myInstance"
 * @param service   Service label  e.g. "#_mysrv#_udp"
 * @param txt       Text Records   e.g. "#name1=value1#name2=value2"
 * @param txtlen    length of txt-string
 * @param port      service port in network byte order
 * @param priority  Host Priority in network byte order (see RFC 2782)
 * @param weight    relative weight for entries with same priority in network
 *                  byte order (see RFC 2782)
 *
 * @note the '#' sign in the above descriptions is to be replaced by the length
 * of the following string. For "#myInstance" you'd have to put 0x0A for '#'.
 *
 * @return ID of registered service. On Error NIP_DNSSD_NO_SERVICE will be
 * returned and nip_error be set accordingly. Possible errors are
 *  - NIP_E_OUT_OF_RESSOURCES
 *  - NIP_E_NOT_CONFIGURED if host name has not yet been configured
 *  - NIP_E_INTERNAL_ERROR if some unexpected error occured
 *  - ...
 */
nip_dnssd_id_t nip_dnssd_register( nip_mem_ptr_t *instance, nip_mem_ptr_t *service, nip_mem_ptr_t *txt, nip_mem_size_t txtlen, uint16_t port, uint16_t priority, uint16_t weight )
{
	nip_dnssd_id_t           res    = NIP_DNSSD_NO_SERVICE;
	nip_dnssd_id_t           new_id = 1;
	nip_error_t              err;
	struct nip_dnssd_service *srv;
	struct nip_mdns_record   rec;
	struct nip_mdns_cache    *cache;
	nip_mem_ptr_t            txt_str_ptr;

	/// @todo Get available service ID
	/// @todo Manage list with more than one items
	srv = &nip_dnssd_services[0];

	// no hostname, try again later
	if  ( nip_mdns_hostname == NIP_MDNS_NO_CACHE_ID )
	{
		nip_error = NIP_E_NOT_CONFIGURED;
		goto quit;
	}

	if ( srv->id != NIP_DNSSD_NO_SERVICE )
	{
		nip_error = NIP_E_OUT_OF_RESSOURCES;
		goto quit;
	}

	// register names
	rec.name.part[0] = NIP_MDNS_REG_STRING( instance);
	if ( rec.name.part[0] == NIP_MDNS_NO_NAMESTRING )
		goto quit;
	NIP_MDNS_REG_NAME( &rec.name.part[1], 2, service );
	if ( rec.name.part[1] == NIP_MDNS_NO_NAMESTRING
	  || rec.name.part[2] == NIP_MDNS_NO_NAMESTRING )
		goto release_name;
	rec.name.part[3] = nip_mdns_domain;
#if NIP_MDNS_MAX_NAME_DEPTH > 4
	rec.name.part[4] = NIP_MDNS_NO_NAMESTRING;
#endif

	// copy TXT records to a managed string
	txt_str_ptr.id  = nip_mem_alloc( txtlen, txtlen, 0, NULL );
	if ( txt_str_ptr.id == NIP_MEM_NULL)
		goto release_name;
	txt_str_ptr.ptr = (void*)0;
	nip_mem_insert( &txt_str_ptr, txt, txtlen );
	// get host name for srv record
	cache = nip_mdns_cache_ptr( nip_mdns_hostname );
	if ( cache == NULL )
	{
		nip_error = NIP_E_INTERNAL_ERROR;
		goto release_name;
	}
	// set srv record data
	nip_memcpy( &rec.data.srv.target, &cache->record.name, sizeof( struct nip_mdns_name ) );
	rec.data.srv.weight   = weight;
	rec.data.srv.priority = priority;
	rec.data.srv.port     = port;
	rec.ttl    = NIP_MDNS_DEFAULT_TTL_A;
	rec.type   = NIP_MDNS_TYPE_SRV;
	nip_mem_release_block( nip_mdns_cache.id );
	// register srv record
	if ( (err = NIP_MDNS_REGISTER( &srv->srv_rr, &rec )) != NIP_E_OK )
	{
		nip_error = err;
		goto release_name;
	}


	// register txt record
	rec.ttl   = NIP_MDNS_DEFAULT_TTL;
	rec.type  = NIP_MDNS_TYPE_TXT;
	rec.data.oth.string = txt_str_ptr.id;
	if ( (err = NIP_MDNS_REGISTER( &srv->txt_rr, &rec )) != NIP_E_OK )
	{
		NIP_MDNS_UNREGISTER( &srv->srv_rr );
		nip_error = err;
		goto release_name;
	}

	// register ptr record
	nip_memcpy( &rec.data.ptr.name, &rec.name, sizeof( struct nip_mdns_name ) );
	rec.name.part[0] = rec.name.part[1];
	rec.name.part[1] = rec.name.part[2];
	rec.name.part[2] = rec.name.part[3];
	rec.name.part[3] = rec.name.part[4];
	rec.ttl   = NIP_MDNS_DEFAULT_TTL_A;
	rec.type  = NIP_MDNS_TYPE_PTR;
	if ( (err = NIP_MDNS_REGISTER( &srv->ptr_rr, &rec )) != NIP_E_OK )
	{
		NIP_MDNS_UNREGISTER( &srv->srv_rr );
		NIP_MDNS_UNREGISTER( &srv->txt_rr );
		nip_error = err;
		goto release_name;
	}

	srv->id = new_id;
	res     = new_id;

	// register service.dns-sd... ptr

	release_name:
		NIP_MDNS_UNREG_NAME( rec.name.part, 3 );

	quit:
		return res;
}

#endif
