/*##############################################################################

nIP - nano IP stack

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

/** @file rtl8019.c
 *
 * Driver for one RTL8019 network interface card, (or probably any other NE2000
 * compatible card).
 *
 * The driver interface is designed to work with the nIP internet stack and
 * relies on its special memory management.
 *
 *@note This driver is written to be used with just one NIC at a time but may be
 * adapted easily to work with several cards simulaneously. For that the static
 * configuration of the address and data ports, the interrupt and the MAC
 * address would have to be moved from rtl8019.h to a dynamically configurable
 * space. Also the global rtl_net_if pointer variable will have to be moved
 * locally into the driver functions and be resolved there using the function
 * nip_net_if_ptr().
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include "nip_init.h"
#include NIP_ARCH_FILE(rtl8019.h)
#include "nip_types.h"
#include "dispatcher.h"
#include "net/net_if.h"
#include "net/ethernet.h"

#if (NIP_RTL_ENABLE == 1)

// mac address
//uint8_t rtl_mac_addr[] = { 0x00, 0x20, 0x18, 0xB1, 0x06, 0x31 };

/** @note this driver is currently written to support one ethernet adapter only.
 * It can easily be adapted for use with several adapters though, by moving the
 * global rtl_net_if pointer variable into all driver functions and have to
 * pointer be resolved locally, depending on the given net_if_id variable. The
 * rtl_next_packet variable would have to be handled on a per-interface basis.*/
nip_net_if_id_t  rtl_net_if_id = NIP_NET_NO_IF;
nip_net_if_t    *rtl_net_if;
uint8_t          rtl_next_packet;

//Routinen zum ansteuern der Netzwerkkarte

//Prototype
void WriteRTL (unsigned char,unsigned char);
unsigned char ReadRTL (unsigned char);
void Write_Ethernet_Init (unsigned int Packet_Length);
void Write_Ethernet_Buffer (unsigned char *buffer, unsigned int bufferlen);
void Write_Ethernet_Terminate (unsigned int bufferlen);
int Read_Ethernet_Init (void);
void Read_Ethernet_Terminate (unsigned int count);
void Read_Ethernet_Byte (unsigned char *buffer, unsigned int count);



/** nIP interface to enable rtl8019 network interface
 * @todo check interface flags. Enable IRQ.
 * @todo set correct return value
 */
nip_success_t nip_rtl_open( nip_net_if_id_t net_if_id )
{
	uint8_t       tmp; /* variable used for all kinds of char-purposes */
	unsigned int  a;
	unsigned long i = 0;
	nip_phy_if_t    *phy_conf;

	// disable AVR interrupt line
	RTL_DISABLE_INT();

	// get network interface configuration
	rtl_net_if = nip_net_if_ptr( net_if_id );
	phy_conf   = &rtl_net_if->phy_conf;


	//Set Addr Port Direction = Output
	DATA_ADDR_RLT = OUTPUT;

	//Set all Addrport High
	ADDR_PORT_RLT = OUTPUT;
	//set Data Port Direction = Input
	DATA_CTRL_RLT = INPUT;

	//Set Reset Pin Low
	RTL_RESET_OFF();

	//added by holgi
	//meine Karten laufen nur wenn man hier noch ca. 100ms wartet
	//wait a long time
	while (i < 60000 * 5) i++;
	//end added by holgi

	RTL_CLEAR_ISR();

	ReadRTL (RTL_RSTPORT);
	WriteRTL(RTL_RSTPORT , 0xFF);


	//wait a short time, at least 1.6ms
	while (a < 1000)
	{
	a++;
	}

	// Put NIC in stop mode
	WriteRTL (RTL_CR , ( RTL_CR_STOP | RTL_CR_RD_ABORT_CMPLTE ));

	// wait until RST bit of status Register is cleared
	do
	{
		tmp = ReadRTL( RTL_ISR );
	} while ( (tmp & 0x80) == 0 );

	// Check Board ID
	tmp = ReadRTL( RTL_8019ID0 ); //should be 0x50
	if ( tmp != 0x50 ) return NIP_ERROR;
	tmp = ReadRTL( RTL_8019ID1 ); //should be 0x70
	if ( tmp != 0x70 ) return NIP_ERROR;



/*	//wait a short time
	a = 0;
	while (a < 1000)
	{
	a++;
	}
*/
	/* ************************************************************************* *
	 * Standard Configuration                                                    *
	 * ************************************************************************* */


	WriteRTL (RTL_DCR   , DCRVAL);
	WriteRTL (RTL_RBCR0 , 0x00);
	WriteRTL (RTL_RBCR1 , 0x00);
#if defined( ETH_NIC_CHECK_MAC )
	WriteRTL (RTL_RCR   , 0xDF/*RCRVAL_CHECK_PHY*/);
#else
	WriteRTL (RTL_RCR   , RCRVAL_IGN_PHY);
#endif

	// set loopback mode
	WriteRTL (RTL_TCR   , 0x02);
	// configure receive buffer ring
	WriteRTL (RTL_TPSR  , RTL_RXSTART);
	WriteRTL (RTL_PSTART, RTL_RXSTART);
	WriteRTL (RTL_BNRY  , RTL_RXSTART);
	WriteRTL (RTL_PSTOP , RTL_RXSTOP);

	// clear interrupt status register
	RTL_CLEAR_ISR();

	/* ************************************************************************* *
	 * Configure interrupt behaviour through Interrupt Mask Register (IMR)       *
	 * ************************************************************************* */

	// add interrupt for packet reception (PRX)
	tmp = RTL_ISR_PRX;

	/* add receive buffer overwrite interrupt (OVW). The buffer may overrun because
	 * we missed previous packet reception interrupts. The OVW interrupt is to make
	 * sure we finally empty the buffer.
	 */
	tmp |= RTL_ISR_OVW;

	// add receive error interrupt (RXE) for logging purpose
	if ( phy_conf->flags & NIP_PHY_IFF_CNT_RXE ) tmp |= RTL_ISR_RXE;

	// add transmit error interrupt (TXE) for logging purpose
	if ( phy_conf->flags & NIP_PHY_IFF_CNT_TXE ) tmp |= RTL_ISR_TXE;

	// write Interrupt Mask Register
	WriteRTL (RTL_CR , ( RTL_CR_STOP  | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE0 ));
	WriteRTL (RTL_IMR , tmp);




	/* ************************************************************************* *
	 * Configure Physical Address                                                *
	 * ************************************************************************* */


	// Read MAC address from Realtek's EEPROM. The EEPROM is mapped
	// to address 0x00 of the Realtek RAM, so that's the address we're trying
	// to read from, using Remote DMA.
	// set remote start address

	WriteRTL( RTL_RSAR0, 0x00 );
	WriteRTL( RTL_RSAR1, 0x00 );

	// Set the Remote DMA Byte Count to 6
	WriteRTL( RTL_RBCR0, 0x0C );
	WriteRTL( RTL_RBCR1, 0x00 );

	// Issue the Remote DMA read command and switch to register page 1 which will be
	// needed to access the PAR registers to store the MAC address in.
	WriteRTL (RTL_CR ,RTL_CR_START | RTL_CR_RD_REMOTE_READ | RTL_CR_SELECT_PAGE1 );

	// Read 6 Bytes and write to Physical-Address-Registers (PAR)
	tmp = 0;
	do{
		// read byte
		phy_conf->hw_addr[tmp] = ReadRTL ( RTL_RDMAPORT );
		phy_conf->hw_brdcst_addr[tmp] = 0xFF;
		// advance to next byte
		ReadRTL ( RTL_RDMAPORT );
#if defined( ETH_PHY_CHECK_MAC )
		// write byte to PAR
		WriteRTL ((RTL_PAR0 + tmp), phy_conf->hw_addr[tmp]);
#endif
		// set multicast mask to receive all multicasts
		/// @todo set multicast mask depending on group membership
		WriteRTL ((RTL_MAR0 + tmp), 0xFF );
	}while (++tmp < 6);


	// Set current page
	rtl_next_packet = RTL_RXSTART ;
	WriteRTL (RTL_CURR , rtl_next_packet );

	// Configure and activate AVR interrupt line
	RTL_CONFIGURE_INT();
	RTL_ENABLE_INT();

	// Put NIC in start mode
	WriteRTL (RTL_CR ,( RTL_CR_START | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE0 ));

	// Remove NIC from loopback mode
	WriteRTL (RTL_TCR , TCRVAL);

	// NIC is now ready to receive and transmit packets

	return NIP_SUCCESS;
}


/** Initialize reading new packat from network interface.
 *
 * Before reading a packet from the RTL8019 buffer, 4 bytes have to be read
 * which include status information and packet length.
 *
 * @return NIP_E_OK or NIP_E_NO_PACKET
 */
nip_success_t nip_rtl_read_init ( nip_net_if_id_t net_if_id )
{
	#if DEBUG_CALLS
		print("---- Read_Ethernet_Init --------------------------\n");
	#endif
	uint8_t tmp, tmp1, sreg;
	nip_phy_if_t    *phy_conf = &rtl_net_if->phy_conf;
	// make sure the following operations aren't interrupted
	sreg = SREG;
	cli();

	// Read Buffer pointer registers. The Boundary Register (BNRY) holds a
	// pointer to the last page that has been read. The Current Page
	// Register (CURR) points to the page address of the first receive
	// buffer page. If both registers are equal, no new packet data is
	// available.

	WriteRTL ( RTL_CR ,( RTL_CR_START | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE1 ));
	tmp1 = ReadRTL(RTL_CURR); //auslesen NIC Register curr

	WriteRTL ( RTL_CR ,( RTL_CR_START | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE0 ));
	tmp = ReadRTL(RTL_BNRY); //auslesen NIC Register bnry

	if ( tmp1 == tmp )
	{
		// queue empty
		SREG = sreg;
		return NIP_ERROR;
	}

	WriteRTL( RTL_RSAR0, 0x00 );
	WriteRTL( RTL_RSAR1, tmp );

	WriteRTL( RTL_RBCR0, 0x04 );
	WriteRTL( RTL_RBCR1, 0x00 );

	WriteRTL ( RTL_CR ,( RTL_CR_START | RTL_CR_RD_REMOTE_READ | RTL_CR_SELECT_PAGE0 ));

	//WriteRTL( RTL_CR ,( RTL_CR_START | RTL_CR_RD_SEND_PACKET ));
	tmp = ReadRTL ( RTL_RDMAPORT ); // content of Receive Status Register (RSR)
	rtl_next_packet = ReadRTL ( RTL_RDMAPORT ); // pointer to next packet in buffer
	tmp = ReadRTL ( RTL_RDMAPORT ); // packet length (LSB)
	tmp1 = ReadRTL( RTL_RDMAPORT ); // packet length (MSB)

	phy_conf->rx_bytes = (tmp1<<8) + tmp - RTL_ETHER_TRAILER_SIZE;
	phy_conf->rx_pos   = 4;

// 	WriteRTL ( RTL_CR ,( RTL_CR_STOP | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE0 ));

	SREG = sreg;
	return NIP_SUCCESS;
}

/** Copy number of bytes from NIC memory into buffer. If buffer is NULL the read
 * bytes will be discarded. That effect is usually used when discarding a packet.
 *
 * You have to call nip_rtl_read_init() once per packet before doing any read
 * operations.
 */
uint16_t nip_rtl_read ( nip_net_if_id_t net_if_id, uint8_t *buffer, uint16_t count)
{
	uint8_t sreg = SREG;
	cli();
	#if DEBUG_NIC
		printf("Read %d Bytes from Network Interface\n",count);
	#endif

	uint16_t ctr;
	nip_phy_if_t    *phy_conf = &rtl_net_if->phy_conf;

	// Jump to set current position in receive buffer

	WriteRTL( RTL_RSAR0, phy_conf->rx_pos & 0x00FF );
	WriteRTL( RTL_RSAR1, ReadRTL(RTL_BNRY) + ( (phy_conf->rx_pos&0xFF00) >> 8 ));

	WriteRTL( RTL_RBCR0, count );
	WriteRTL( RTL_RBCR1, count >> 8 );

	WriteRTL ( RTL_CR ,( RTL_CR_START | RTL_CR_RD_REMOTE_READ | RTL_CR_SELECT_PAGE0 ));

	//Der Datenport wir auf input gesetzt vom NIC zum IC lesen
	DATA_CTRL_RLT = INPUT;
	//write Adress to Port
	ADDR_PORT_RLT = RTL_RDMAPORT;

	//Daten in den Buffer uebernehmen
	for(ctr=0; ctr<count && phy_conf->rx_bytes > 0; ctr++)
	{
		//set read Pin to Low
		RTL_RD_OFF();
		//wait a short time
		nop();
		nop();
		//read from port rtldata
		if ( buffer != NULL )
			buffer[ctr] = DATA_PORT_RLT_READ;
		//set Read Pin to High
		RTL_RD_ON();
		//read from RTL complete

		//decrement packet byte count
		phy_conf->rx_bytes--;

		// packet has been read completely
		if ( phy_conf->rx_bytes == 0 )
		{
			// Update boundary register. This should be done after having read a
			// packet which we assume to have happened after all bytes were read.
			if ( rtl_next_packet < RTL_RXSTART )
				WriteRTL (RTL_BNRY  , RTL_RXSTOP - 1 );
			else
				WriteRTL (RTL_BNRY  , rtl_next_packet );

			ctr++;
			break;
		}
	}
	phy_conf->rx_pos += ctr;
// 	WriteRTL ( RTL_CR ,( RTL_CR_STOP | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE0 ));

	SREG = sreg;

	return ctr;
}

/** Initialize transmission of Ethernet packet. If all header data is complete
 *  the header will be written to packet buffer. If daddr is NULL, ARP will be
 *  called to initiate address resolution.
 *  @param net_if_id
 *  @param ll_target_addr  ip address of local link target
 *  @param type            ethernet type
 *  @param daddr           pointer to array with destination MAC address
 *  @param saddr           pointer to array with source MAC address
 *  @param len             size of packet to be sent
 *  */
nip_error_t nip_rtl_send_init( nip_net_if_id_t net_if_id, void *ll_target_addr,
uint16_t type, void *daddr, void *saddr, uint16_t len)
{
	nip_phy_if_t    *phy_conf = &rtl_net_if->phy_conf;
	uint8_t         sreg;
	// if there are untransmitted bytes, assume the device is busy transmitting
	// another packet.
	if ( phy_conf->tx_bytes > 0 )
		return NIP_E_DEV_BUSY;

	// make sure the following operations aren't interrupted
	sreg = SREG;
	cli();

	len += 14; // add ethernet header length

	// configure Ethernet transmission
	phy_conf->tx_bytes = len ;
	phy_conf->tx_pos   = 0 ;
	if ( len < RTL_ETHER_MIN_SIZE )
		len = RTL_ETHER_MIN_SIZE;

	WriteRTL ( RTL_TPSR , RTL_TXSTART );
/*	WriteRTL ( RTL_RSAR0, 0x00   );
	WriteRTL ( RTL_RSAR1, RTL_TXSTART );
//	WriteRTL ( RTL_ISR  , (1<<PRX|1<<PTX|1<<RXE|1<<TXE|1<<OVW|1<<CNT|1<<RDC|1<<RST));
	WriteRTL ( RTL_RBCR0, ( len & 0x00ff ));
	WriteRTL ( RTL_RBCR1, ( len & 0xff00 ) >>8 );*/

//	WriteRTL ( RTL_CR   , ( RTL_CR_START | RTL_CR_RD_REMOTE_WRITE ));
	WriteRTL ( RTL_TBCR0, ( len & 0x00ff ) );
	WriteRTL ( RTL_TBCR1, ( len & 0xff00 ) >>8 );

	// Build and send Ethernet Header
	nip_rtl_send( net_if_id, daddr, 6 );
	nip_rtl_send( net_if_id, saddr, 6 );
	nip_rtl_send( net_if_id, (uint8_t*)&type , 2 );

	SREG = sreg;

	return NIP_E_OK;
}

uint16_t nip_rtl_send( nip_net_if_id_t net_if_id, uint8_t* buffer, uint16_t count)
{
	nip_phy_if_t    *phy_conf = &rtl_net_if->phy_conf;
	uint16_t ctr = phy_conf->tx_pos + phy_conf->tx_bytes;
	uint8_t         sreg;

	// make sure the following operations aren't interrupted
	sreg = SREG;
	cli();

	if ( ctr < RTL_ETHER_MIN_SIZE)
		ctr = RTL_ETHER_MIN_SIZE;

	WriteRTL ( RTL_CR   , ( RTL_CR_STOP | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE0 ));
// 	WriteRTL ( RTL_CR   , ( RTL_CR_START | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE0 ));

	WriteRTL ( RTL_RSAR0, (phy_conf->tx_pos & 0x00FF) );
	WriteRTL ( RTL_RSAR1, RTL_TXSTART + ((phy_conf->tx_pos&0xFF00)>>8));
	WriteRTL ( RTL_RBCR0, ( ctr & 0x00ff ) );
	WriteRTL ( RTL_RBCR1, ( ctr & 0xff00 ) >>8 );

	WriteRTL ( RTL_CR   , ( RTL_CR_START | RTL_CR_RD_REMOTE_WRITE | RTL_CR_SELECT_PAGE0 ));

	//Der Datenport wir auf output gesetzt vom IC zum NIC schreiben
	DATA_CTRL_RLT = OUTPUT;
	//write Adress to Port
	ADDR_PORT_RLT = RTL_RDMAPORT;

	for ( ctr = 0; ; )
	{
		// break, if count bytes have been sent and end of transmission has not
		// been reached
		if ( ctr == count && phy_conf->tx_bytes != 0 ) break;

		// break, if end of transmission has been reached and no padding is
		// necessary
		if ( phy_conf->tx_bytes == 0 && ++phy_conf->tx_pos >= RTL_ETHER_MIN_SIZE ) break;

		// write data to DMA channel
		if ( phy_conf->tx_bytes > 0 )
		{
			phy_conf->tx_bytes--;
			// write Buffer-Data to Port
			DATA_PORT_RLT_WRITE = buffer[ctr];
			ctr++;
		} else
		{
			// write Padding to Port
			DATA_PORT_RLT_WRITE = 0x00;
		}

		//set write Pin to Low
		RTL_WR_OFF();
		//wait a short time
        	nop();
        	nop();

		//set write Pin to High
		RTL_WR_ON();
		//write to RTL complete

	}

	phy_conf->tx_pos += ctr;

	//loop_until_bit_is_set(ReadRTL(isr),rdc);
	///@todo timeout might be needed in rare cases (if transmission failed)
	///@todo check if the following line is needed
	//while ((ReadRTL(RTL_ISR) & RTL_ISR_RDC) == 0) {}

	// terminate session if necessary
	if ( phy_conf->tx_bytes == 0 )
	{

		WriteRTL (RTL_CR,( RTL_CR_TRANSMIT_PACKET | RTL_CR_RD_ABORT_CMPLTE ));

	}
// 	WriteRTL ( RTL_CR   , ( RTL_CR_STOP | RTL_CR_RD_ABORT_CMPLTE | RTL_CR_SELECT_PAGE0 ));

	SREG = sreg;
	return ctr;
}


/** nIP interface to disable rtl8019 network interface
 * @todo disable network interface and irq.
 * @todo set correct return value
 */
nip_success_t nip_rtl_close( nip_net_if_id_t net_if_id )
{

	return NIP_SUCCESS;
}


#ifdef RTL_INTERRUPT

/** Interrupt Service Routine to handle interrupt requests from network interface.
 *
 * @todo Unterbrechbare Interrupts verwenden: void RTL_INTERRUPT (void)__attribute__((interrupt));
 *
 */

ISR( RTL_INTERRUPT )
{
	uint8_t isr = 0;

	// don't handle interrupt if network interface has not been configured
	if ( rtl_net_if_id == NIP_NET_NO_IF ) return;

	// read Interrupt Status Register (ISR)
	isr = ReadRTL(RTL_ISR);

	// check receive buffer overwrite
	if ( isr & RTL_ISR_OVW )
	{
		/// @todo Handle overwrites according to developer guide
		NIP_PHY_LOG_OVW( rtl_net_if );
	}

	// check for receive errors
	if ( isr & RTL_ISR_RXE ) NIP_PHY_LOG_RXE( rtl_net_if );

	// check for transmit errors
	if ( isr & RTL_ISR_TXE ) NIP_PHY_LOG_TXE( rtl_net_if );

	// check for reset interrupt
	if ( isr & RTL_ISR_RST )
	{
		// TODO do something?
	}

	// check for packet reception
	if ( isr & RTL_ISR_PRX )
	{
		// notify dispatcher and execute if not already running
		nip_disp_notify_if( rtl_net_if_id, NIP_DISP_IF_ETH_R );
// 		nip_dispatcher( );
	}


	//Zuruecksetzen der Interrupt Bits des NIC
	RTL_CLEAR_ISR();

}

#endif /* RTL_INTERRUPT*/


//############################################################################
// Schreibt ein Byte in ein angegebenes Register der Netzwerkkarte
void WriteRTL (unsigned char rtl_addr,unsigned char rtl_data)
//############################################################################
{
	//Der Datenport wird auf output gesetzt vom IC zum NIC schreiben
	DATA_CTRL_RLT = OUTPUT;
	//write Adress to Port
	ADDR_PORT_RLT = rtl_addr;
	//write Data to Port
	DATA_PORT_RLT_WRITE = rtl_data;
	//set write Pin to Low
	RTL_WR_OFF();
	//wait a short time
        nop(); //gib mal ein bisschen GAS Ulrich ;)
        nop();

	//set write Pin to High
	RTL_WR_ON();
	//write to RTL complete
}


//############################################################################
// Auslesen eines Byte von einen angegebenes Register der Netzwerkkarte
unsigned char ReadRTL (unsigned char rtl_addr)
//############################################################################
{
	unsigned char rtl_data;
	//Der Datenport wir auf input gesetzt vom NIC zum IC lesen
	DATA_CTRL_RLT = INPUT;
	//write Adress to Port
	ADDR_PORT_RLT = rtl_addr;
	//set read Pin to Low
	RTL_RD_OFF();
	//wait a short time
        nop();
        nop();

	//read from port rtldata
	rtl_data = DATA_PORT_RLT_READ;
	//set Read Pin to High
	RTL_RD_ON();
	//read from RTL complete
	return (rtl_data);
}


/*
//############################################################################
void Write_Ethernet_Terminate (unsigned int bufferlen)
//############################################################################
{
	#if DEBUG_CALLS
		print("---- Write_Ethernet_Terminate --------------------\n");
	#endif

	#if DEBUG_NIC
		printf("Padding: %d Bytes",
		        bufferlen<RTL_ETHER_MIN_SIZE ? RTL_ETHER_MIN_SIZE-bufferlen : 0);
	#endif

	unsigned char ctr;
	if (bufferlen < RTL_ETHER_MIN_SIZE)
	{
		for (ctr = bufferlen; ctr < RTL_ETHER_MIN_SIZE; ctr++){
			//write Data to Port
			DATA_PORT_RLT_WRITE = 0x00;
			#if DEBUG_NIC && DEBUG_VERBOSE
				printf("__");
			#endif
			//set write Pin to Low
			RTL_WR_OFF();
			//wait a short time
			nop(); //gib mal ein bisschen GAS Ulrich ;) Holgi
			nop();

			//set write Pin to High
			RTL_WR_ON();
			//write to RTL complete
		}
	}
	#if DEBUG_NIC
		print(".\n");
	#endif

	//loop_until_bit_is_set(ReadRTL(isr),rdc);
	while ((ReadRTL(RTL_ISR) & RTL_ISR_RDC) == 0) {}

	if (bufferlen<RTL_ETHER_MIN_SIZE) bufferlen = RTL_ETHER_MIN_SIZE;

	WriteRTL (RTL_TBCR0,(bufferlen & 0x00ff));
	WriteRTL (RTL_TBCR1,(bufferlen & 0xff00) >>8);
	WriteRTL (RTL_CR,( RTL_CR_TRANSMIT_PACKET | RTL_CR_RD_ABORT_CMPLTE ));



	//Globale Interrupts einschalten
//	sei();
}
*/

/*
//############################################################################
void Read_Ethernet_Terminate (unsigned int count)
//############################################################################

{
	#if DEBUG_CALLS
		print("---- Read_Ethernet_Terminate ---------------------\n");
	#endif
	unsigned int ctr;
	unsigned char tmp = 0;
	#if DEBUG_NIC
		printf("Drop remaining %d Bytes and terminate reading.\n",count);
	#endif

	for(ctr=0; ctr<(count+RTL_ETHER_TRAILER_SIZE); ctr++)
	{

		//set read Pin to Low
		RTL_RD_OFF();
		//wait a short time
		nop();
		nop();
		//read from port rtldata
		tmp = DATA_PORT_RLT_READ;
		//set Read Pin to High
		RTL_RD_ON();
		//read from RTL complete

		#if DEBUG_NIC && DEBUG_VERBOSE
			printf("%02x", tmp);
		#endif

	}
	#if DEBUG_NIC && DEBUG_VERBOSE
		print("\n");
	#endif

	tmp = tmp & RTL_ISR_RDC;

	if ( tmp != 64 )
	{
		ReadRTL(RTL_ISR);
	}

}*/

#endif  /* (NIP_RTL_ENABLE == 1) */
