/*##############################################################################

nIP - nano IP stack

File        : ethernet.c

Description : Functions to serve ethernet interfaces

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

##############################################################################*/

#include "dispatcher.h"
#include "net/ethernet.h"
#include "net/arp.h"
#include "net/ip.h"

#if (NIP_ETH_ENABLE == 1)

// string compare function only needed for mac address check
#if !defined( ETH_NIC_CHECK_MAC )
#	include <string.h>
#endif

#include "nip_init.h"
#include "net/net_if.h"

extern nip_net_if_t nip_net_if_list[];
extern nip_disp_state_t nip_disp;


/** fill ethernet interface configuration with default values
 *
 * @param id  id of network interface to configure
 **/
void nip_eth_configure_if_defaults( nip_net_if_id_t id )
{
	nip_net_if_list[id].phy_conf.flags = NIP_PHY_IFF_BROADCAST;
	nip_net_if_list[id].phy_conf.hw_addr_size = 6;
	nip_net_if_list[id].phy_conf.open  = NULL; /* driver will fill in */
	nip_net_if_list[id].phy_conf.close = NULL; /* driver will fill in */
	nip_net_if_list[id].phy_conf.hard_disp_recv = &nip_eth_disp_receive;
	nip_net_if_list[id].phy_conf.hard_send_init = NULL; /* driver will fill in */
	nip_net_if_list[id].phy_conf.hard_send      = NULL; /* driver will fill in */
	nip_net_if_list[id].phy_conf.hard_read_init = NULL; /* driver will fill in */
	nip_net_if_list[id].phy_conf.hard_read      = NULL; /* driver will fill in */
}


/** Function to be called upon packet reception from Network Interface.
 *
 * The function will read the first packet from buffer and hand it to the
 * appropriate protocol handler. No decision on discarding the packet will be
 * made here, unless the ethernet address check fails. Discarding due to lack
 * of ressources is up to the protocol handler which may decide to have no
 * buffer space for the data. Also it's up to the protocol handler to decide
 * which transmission the packet will be assigned to (e.g. only IP can know if
 * the received data belongs to a fragmented packet waiting to be completed).
 */
void nip_eth_disp_receive( void )
{
	uint8_t tmp[6];
	uint8_t type[2];
	nip_net_if_id_t tmp_if_id = NIP_CURR_IF;
#if !defined( ETH_NIC_CHECK_MAC )
	uint8_t tmp1;
#endif
	// Packet reception should be completed at this point. There may be packets
	// waiting in the interface's buffer, but the previous packet should've been
	// received completely. To avoid errors and deadlocks this routine double
	// checks on that and empties the receive-buffer if necessary.
	if ( nip_net_if_list[ tmp_if_id ].phy_conf.rx_bytes > 0 )
	{
		nip_net_if_list[ tmp_if_id ].phy_conf.hard_read(
			NIP_CURR_IF,
			NULL,
			nip_net_if_list[ tmp_if_id ].phy_conf.rx_bytes
		);
	}


	// initialize packet reading
	if ( nip_net_if_list[tmp_if_id].phy_conf.hard_read_init( tmp_if_id ) != NIP_E_OK )
		return;

	// make sure we come back and check for another packet (we don't know how
	// many packets have been received)
 	nip_disp_notify_if( tmp_if_id, NIP_DISP_IF_ETH_R );

	// read destination address
	nip_net_if_list[tmp_if_id].phy_conf.hard_read( tmp_if_id, tmp, 6 );

// if the NIC is not doing the physical address check, we'll have to do that
#if !defined( ETH_NIC_CHECK_MAC )
	// Check if destination address is broadcast address
	tmp1 = 0;
	do
	{
		if ( tmp[tmp1] != 255 ) break;
	}
	while ( ++tmp1 < 6 );

	// If all 6 Bytes equalled 255, the given message is a ethernet broadcast
	if (  tmp1 == 6 )
	{
	}
	else
#endif
	// If the least significant bit of the first address byte equals 1,
	// the given ethernet address is a (multicast) group address.
	if ( (tmp[0]& 0x01) > 0 )
	{
		 // TODO check multicast destination. Is this message for us? Treat
		 // broadcast as special kind of multicast. Remove above broadcast check
		 // when done.
		 // if check failed goto terminate_read
	}

// if the NIC is not doing the physical address check, we'll have to do that
#if !defined( ETH_NIC_CHECK_MAC )
	else
	// check unicast destination
	if ( nip_memcmp( tmp, nip_net_if_list[tmp_if_id].phy_conf.hw_addr, 6 ) == 0 )
	{
	}
	else
	{
		// Ethernet address check failed
		// TODO discard packet

		// we use 'goto' here because it's the only way to skip the following
		// lines without having to use additional status variables and conditions.
		// (That saves about 32 Bytes of lovely program memory over the next
		// favorable solution of using tmp1 as status variable indicating wether
		// we're the the correct receiver of the packet)
		goto terminate_read;
	}
#endif

	// read source address
	nip_net_if_list[tmp_if_id].phy_conf.hard_read( tmp_if_id, nip_disp.next.common.remote_hw_addr, 6 );

	// read type
	nip_net_if_list[tmp_if_id].phy_conf.hard_read( tmp_if_id, type, 2 );

	// eval type
	#if (NIP_ARP_ENABLE == 1)
	if ( type[0] == 0x08 && type[1] == 0x06 )
	{
		// received ARP packet
		NIP_CURR_CMD = &nip_disp_arp_receive;;
		return;
	}
	#endif
	#if (NIP_IP_ENABLE == 1)
	if ( type[0] == 0x08 && type[1] == 0x00 )
	{
		// received IP packet
		NIP_CURR_CMD = &nip_ip_disp_receive;;
		return;
	}
	#endif

#if !defined( ETH_NIC_CHECK_MAC )
	terminate_read:
	// read remaining packet bytes from NIC
	nip_net_if_list[tmp_if_id].phy_conf.hard_read(
		tmp_if_id,
		NULL,
		nip_net_if_list[tmp_if_id].phy_conf.rx_bytes
	);
#endif

	return;
}

#endif /* NIP_ETH_ENABLE == 1 */
