/*##############################################################################

nIP - nano IP stack

File        : tcp.c

Description : TCP

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#############################################################################*/

#include "nip_init.h"
#include "dispatcher.h"
#include "inet.h"
#include "os_core.h"
#include "nip_error.h"
#include "net/net_if.h"
#include "net/tcp.h"
#include "net/numbers.h"

#if NIP_TCP_ENABLE == 1

// static socket array
#if NIP_TCP_MAX_SOCKETS == 1
	#define sockid 0
	struct nip_tcp_sock nip_tcp_sock = 0;
	#define NIP_TCP_SOCKATTR(sockid, attr) nip_tcp_sock.attr
#else
	struct nip_tcp_sock nip_tcp_sockets[ NIP_TCP_MAX_SOCKETS ];
	#define NIP_TCP_SOCKATTR(sockid, attr) nip_tcp_sockets[sockid].attr
#endif

nip_tcp_sock_cnt_t nip_tcp_unaccepted_cnt;
nip_tcp_sock_id_t  nip_tcp_unaccepted[ NIP_TCP_MAX_UNACCEPTED ];
nip_tcp_sock_cnt_t nip_tcp_conn_cnt; /**< number of connections in nip_tcp_conns */
nip_mem_handle_t   nip_tcp_conns = NIP_MEM_NULL; /**< dynamic memory for connection array */


/** Register TCP socket with given address and flags.
 * @param addr structure to be filled with address and port. If all address
 *             bytes are set to Zero, the socket will listen on all interfaces.
 *             If port is set to Zero, the implementation will choose a random
 *             port, which should be the usual way to go for sockets with the
 *             NIP_TCP_SOCK_FLG_LISTENING-flag set.
 * @param flags possible flags are:
 *               - NIP_TCP_SOCK_FLG_NON_BLOCK
 *               - NIP_TCP_SOCK_FLG_LISTENING
 * @return NIP_TCP_NO_SOCKET on error. Global nip_error variable will be set to
 * NIP_E_OUT_OF_RESSOURCES in that case.
 */
nip_tcp_sock_id_t nip_tcp_socket( struct nip_tcp_sock_addr *addr, uint8_t flags )
{
#if NIP_TCP_MAX_SOCKETS > 1
	// more than one socket available?
	// -> Loop through them until we find unused one.
	nip_tcp_sock_id_t sockid = 0;
	do
 	{
#endif
		// check if socket is unused
		if ( (NIP_TCP_SOCKATTR(sockid,flags) & NIP_TCP_SOCK_FLG_USED) == 0 )
		{
			// configure socket
			NIP_TCP_SOCKATTR(sockid,flags) = flags | NIP_TCP_SOCK_FLG_USED;
			nip_memcpy( &NIP_TCP_SOCKATTR(sockid,addr), addr, 4 );
			return sockid;
		}
#if NIP_TCP_MAX_SOCKETS > 1
	} while ( ++sockid < NIP_TCP_MAX_SOCKETS );
#endif

	// out of ressources
	nip_error = NIP_E_OUT_OF_RESSOURCES;
	return NIP_TCP_NO_SOCKET;
}


/** Remove connection TCB and entry in socket list.
 */
void nip_tcp_del_conn( nip_tcp_sock_id_t connid )
{
	struct nip_tcb     *tcb;
	nip_tcp_sock_cnt_t cnt = 0;
	nip_tcp_sock_id_t  sockid = NIP_TCP_NO_SOCKET;
	uint8_t            moveup;
	uint8_t            children;

	if ( connid < NIP_TCP_MAX_SOCKETS )
		sockid = connid;

	do
	{
		moveup = 0;
		children = 0;
		if ( nip_tcp_conn_cnt > 0 )
		{
		// get pointer to tcp array
			tcb = nip_mem_obtain_ptr( nip_tcp_conns );
			if ( tcb != NULL )
			{
				do
				{
					if ( tcb->connid == connid )
					{
						// free buffers
						nip_mem_free( tcb->snd_buf );
						nip_mem_free( tcb->rcv_buf );
						// mark connection for removal
						moveup++;
						sockid = tcb->sockid;
					}
					else
					{
						if ( tcb->sockid == sockid )
							children++;
						// move connection
						nip_memcpy( tcb - moveup, tcb, sizeof( struct nip_tcb ) );
					}
					tcb++;
				} while ( ++cnt < nip_tcp_conn_cnt );

				// reduce connection count
				nip_tcp_conn_cnt -= moveup;

				// free pointer
				nip_mem_release_block( nip_tcp_conns );
				nip_mem_set_used( nip_tcp_conns, nip_tcp_conn_cnt * sizeof( struct nip_tcb ) );
			}
		}
	}
	// repeat loop to count children
	while ( moveup > 0 );

	// Delete socket, if set to autoclose, which means the user confirmed it will
	// not use the socket ID for reading or writing anymore.
	if ( sockid != NIP_TCP_NO_SOCKET
     && NIP_TCP_SOCKATTR(sockid,flags) & NIP_TCP_SOCK_FLG_AUTOCLOSE )
	{
		// Make sure there aren't any spawned child-sockets still using it. If so,
		// flag socket to not accept new connections.
		if ( children > 0 )
		{
			NIP_TCP_SOCKATTR(sockid,flags) &= ~NIP_TCP_SOCK_FLG_LISTENING;
		}
		else
			NIP_TCP_SOCKATTR(sockid,flags) = 0;
	}

}

/** Find transmission control block (tcb) for connection with matching address.
 * If rem_ip is not NULL the given sockid will be matched against the Ids of the
 * socket-array, otherwise sockid will be matched against connection IDs.
 *
 * @return pointer to tcb or NULL if tcb was not found. If NULL is returned,
 * nip_error will contain one of the following error codes
 *  - NIP_E_AGAIN         needed ressources were locked
 *  - NIP_E_INVALID_SOCK  given sockid does not belong to any open connection
 * @note You _have_ to release the given pointer after using it by calling
 *    nip_mem_release_block( nip_tcp_conns );
 */
struct nip_tcb *
nip_tcp_find_conn( nip_tcp_sock_id_t sockid, uint8_t *rem_ip, uint16_t rem_port)
{
	struct nip_tcb     *tcb;
	nip_tcp_sock_cnt_t cnt = 0;
	if ( nip_tcp_conn_cnt > 0 )
	{
		// get pointer to tcp array
		tcb = nip_mem_obtain_ptr( nip_tcp_conns );
		if ( tcb != NULL )
		{
			do
			{
				// Search by address won't retrieve CLOSED sockets.
				if(( tcb->sockid        == sockid
				  && rem_ip             != NULL
				  && tcb->rem_addr.port == rem_port
				  && nip_memcmp( tcb->rem_addr.ip, rem_ip, 4 ) == 0
				  && tcb->status        != NIP_TCP_STAT_CLOSED )
				|| ( tcb->connid  == sockid
				  && rem_ip       == NULL )
				  )
				{
					return tcb;
				}
				tcb++;
			} while ( ++cnt < nip_tcp_conn_cnt );

			// connection not found -> free pointer
			nip_mem_release_block( nip_tcp_conns );
			nip_error = NIP_E_INVALID_SOCK;
		}
		else
		{
			nip_error = NIP_E_AGAIN;
		}
	}

	return NULL;
}

/** Create new transmission control block (tcb) with given parameters
 * @return pointer to tcb or NULL if there were no ressources to setup connection.
 * @note You _have_ to release the given pointer after using it by calling
 *    nip_mem_release_block( nip_tcp_conns );
 */
struct nip_tcb *
nip_tcp_create_conn( nip_tcp_sock_id_t sockid, uint8_t *rem_ip, uint16_t rem_port,
                     nip_tcp_stat_t init_status )
{
	struct nip_tcb     *tcb   = NULL;
	nip_tcp_sock_cnt_t cnt;
	nip_tcp_sock_id_t  connid;
	if ( nip_tcp_conns == NIP_MEM_NULL )
	{
		nip_tcp_conn_cnt = 0;
		nip_tcp_conns = nip_mem_alloc( sizeof( struct nip_tcb ), 0, 0, NULL );
	}

	// append 0-initialized memory for new tcb
	if ( nip_mem_write( nip_tcp_conns, NULL, sizeof( struct nip_tcb ) )
			== NIP_E_OK )
	{
		// get pointer to tcp array
		tcb = nip_mem_obtain_ptr( nip_tcp_conns );
		if ( tcb != NULL )
		{
			// For listening sockets a new connection ID has to be determined, as
			// they are forked off of the current listening socket. Connection Ids
			// start beyond space for normal sockets.
			if ( init_status == NIP_TCP_STAT_LISTEN )
				connid = NIP_TCP_MAX_SOCKETS;
			else
				connid = sockid;

			// check uniqueness of connection-id
			unique_check:
			cnt = 0;
			while ( cnt < nip_tcp_conn_cnt )
			{
				// check if other connection is using that id
				if ( tcb->connid == connid )
				{
					// try new id and restart probing
					connid++;
					tcb -= cnt;
					goto unique_check;
				}
				tcb++;
				cnt++;
			};

			nip_memset( tcb, 0, sizeof( struct nip_tcb ) );

			tcb->snd_buf     = nip_mem_alloc( 0, 0, 0, NULL );
			tcb->rcv_buf     = nip_mem_alloc( 0, 0, 0, NULL );

			// if all went well, return tcb-pointer
			if ( tcb->snd_buf != NIP_MEM_NULL && tcb->rcv_buf != NIP_MEM_NULL )
			{
				// increase connection count
				nip_tcp_conn_cnt++ ;

				// configure tcb
				tcb->sockid = sockid;
				tcb->connid = connid;
				tcb->status = init_status;
				nip_memcpy( tcb->rem_addr.ip, rem_ip, 4 );
				tcb->rem_addr.port = rem_port;

				// initial send sequence number (iss) should be random.. we'll
				// postpone the randomness and base it on the time.
				///@todo make iss random!
				tcb->iss     = nip_tickcount();
				tcb->rcv_wnd = NIP_TCP_DEFAULT_RCV_WND;

				tcb->snd_nxt = tcb->iss + 1; // Next received ACK has to be iss+1
				tcb->snd_una = tcb->iss;

				return tcb;
			}

			// on error free buffers and tcb-pointer
			nip_mem_free( tcb->snd_buf );
			nip_mem_free( tcb->rcv_buf );
			nip_mem_release_block( nip_tcp_conns );

		}

		// failed to set up connection?
		// -> cancel operation, reset size of tcb-array which will clip off the
		// newly created tcb.
		nip_mem_set_used( nip_tcp_conns, nip_tcp_conn_cnt * sizeof( struct nip_tcb ) );
	}


	return NULL;
}

/** Close connection or remove socket if not connected.
*/
void nip_tcp_close( nip_tcp_sock_id_t connid )
{
	struct nip_tcb  *tcb;

	// Stack functions operating outside the dispatcher must not be interrupted
// 	nip_int_off();

	// get pointer to connection TCB
	tcb = nip_tcp_find_conn( connid, NULL, 0 );

	// shutdown mode NIP_TCP_SOCK_FLG_SHUT_WR will actually lead to TCP doing
	// something. NIP_TCP_SOCK_FLG_SHUT_RD is just for the connection to know,
	// that the user won't perform read on it anymore.
	if (tcb == NULL && nip_error == NIP_E_INVALID_SOCK)
	{
		// no open connection found for connid?
		// -> Try to remove connection data. Socket will be removed unless
		// there are still connections depending on it.
		nip_tcp_del_conn( connid );
	}
	else if ( tcb != NULL )
	{
		// Close connection

		if ( tcb->status == NIP_TCP_STAT_SYN_SENT
			|| tcb->status == NIP_TCP_STAT_LISTEN )
		{
			tcb->status = NIP_TCP_STAT_CLOSED;
		}

		if ( tcb->status == NIP_TCP_STAT_SYN_RECEIVED
		|| tcb->status == NIP_TCP_STAT_ESTABLISHED )
		{
			tcb->status = NIP_TCP_STAT_FIN_WAIT1;
		}
		else
		if ( tcb->status == NIP_TCP_STAT_CLOSE_WAIT )
		{
			tcb->status = NIP_TCP_STAT_LAST_ACK;
		}
		else
			goto release;

		// configure transmission for FIN to be sent
		if ( nip_mem_buf_used( tcb->snd_buf ) == 0 )
		{
			tcb->snd_nxt ++;
			tcb->flags   &= ~NIP_TCP_SOCK_FLG_SEND;
		}
	}

	if ( tcb != NULL )
	{
		release:
		tcb->flags |= NIP_TCP_SHUT_WR;
		nip_mem_release_block( nip_tcp_conns );

		// notify TCP to check on state changes
		nip_disp_notify( NIP_DISP_CHECK_TCP );
// 		nip_int_on();

		nip_dispatcher();
	}

// 	nip_int_on();
}

/** Looks for incoming connections on given listening socket.
 *
 */
nip_tcp_sock_id_t nip_tcp_accept( nip_tcp_sock_id_t sockid )
{
	nip_tcp_sock_cnt_t i;
	nip_tcp_sock_id_t  res = NIP_TCP_NO_SOCKET;
	struct nip_tcb     *tcb;

	// Stack functions operating outside the dispatcher must not be interrupted
// 	nip_int_off();

	for ( i = 0; i < nip_tcp_unaccepted_cnt; i++ )
	{
		tcb = nip_tcp_find_conn( nip_tcp_unaccepted[i], NULL, 0 );
		if ( tcb == NULL || (tcb->flags&NIP_TCP_SOCK_FLG_ACCEPTED) )
		{
			// invalid list entry? -> remove
		}
		else if ( tcb->sockid == sockid )
		{
			// found connection
			res = tcb->connid;
			tcb->flags |= NIP_TCP_SOCK_FLG_ACCEPTED;
		}
		else
			goto next;

		// remove entry from list
		nip_tcp_unaccepted_cnt--;
		nip_memmove( &nip_tcp_unaccepted[i],&nip_tcp_unaccepted[i+1],nip_tcp_unaccepted_cnt-i);

		next:
		if ( tcb != NULL )
			nip_mem_release_block( nip_tcp_conns );

		if ( res != NIP_TCP_NO_SOCKET )
			break;
	}

// 	nip_int_on();

	return res;
}

/** Write data to connection
 * @return number of bytes written. If 0 was returned, nip_error will contain one
 * of the following error codes
 *  - NIP_E_EOF    socket has been closed and is now read-only
 *  - NIP_E_AGAIN  not enough space in send buffer or ressource locked
 *  - NIP_E_INVALID_SOCK given sockid did not match any open connection
 */
nip_mem_size_t nip_tcp_write( nip_tcp_sock_id_t connid, uint8_t *buf, nip_mem_size_t size )
{
	struct nip_tcb   *tcb;
	nip_mem_handle_t snd_buf;


	// Stack functions operating outside the dispatcher must not be interrupted
// 	nip_int_off();

	// get pointer to connection TCB
	tcb = nip_tcp_find_conn( connid, NULL, 0 );
	if ( tcb == NULL )
	{
		// nip_error will be set by nip_tcp_find_conn
		size = 0;
	}

	else
	{
		// connection already closed?
		if ( tcb->status == NIP_TCP_STAT_FIN_WAIT1
		|| tcb->status == NIP_TCP_STAT_FIN_WAIT2
		|| tcb->status == NIP_TCP_STAT_CLOSING
		|| tcb->status == NIP_TCP_STAT_TIMEWAIT
		|| tcb->status == NIP_TCP_STAT_LAST_ACK
			)
		{
			nip_error = NIP_E_EOF;
			size = 0;
		}
		snd_buf = tcb->snd_buf;
		nip_mem_release_block( nip_tcp_conns );

		if ( size > 0 && NIP_E_OK != nip_mem_write( snd_buf, buf, size ) )
		{
			nip_error = NIP_E_AGAIN;
			size = 0;
		}
	}

// 	nip_int_on();

	if ( size > 0 )
	{
		// notify TCP to transmit data
		nip_disp_notify( NIP_DISP_CHECK_TCP );
		nip_dispatcher();
	}

	return size;
}

/** Read data from connection into given buffer
 * @return number of bytes copied. If 0 was returned, nip_error will contain one
 * of the following error codes
 *  - NIP_E_EOF    socket has been closed by remote peer
 *  - NIP_E_AGAIN  no data available or ressource locked
 *  - NIP_E_INVALID_SOCK given sockid did not match any open connection
 */
nip_mem_size_t nip_tcp_read( nip_tcp_sock_id_t connid, uint8_t *buf, nip_mem_size_t size )
{
	struct nip_tcb *tcb;

	// Stack functions operating outside the dispatcher must not be interrupted
// 	nip_int_off();

	// get pointer to connection TCB
	tcb = nip_tcp_find_conn( connid, NULL, 0 );
	if ( tcb == NULL )
	{
		// nip_error will be set by nip_tcp_find_conn
		size = 0;
	}
	else
	{
		// read data
		size = nip_mem_read( tcb->rcv_buf, buf, size );

		// read all data from block?
		if ( size == 0 && nip_error == NIP_E_EOF )
		{
			// if remote peer didn't close connection, try again later.
			if ( tcb->status < NIP_TCP_STAT_ESTABLISHED || (tcb->flags & NIP_TCP_SOCK_FLG_RECV) )
			{
				nip_error = NIP_E_AGAIN;
			}

			// socket closed by remote peer? Record this as "user has been notified"
			else
			{
// 				nip_error = NIP_E_EOF;
				tcb->flags |= NIP_TCP_SHUT_RD;
			}
		}
		nip_mem_release_block( nip_tcp_conns );
	}

// 	nip_int_on();

	return size;
}

/** Calculate TCP checksum. If trans->header is set the checksum will be
 * inserted there. Make sure neither trans->header nor trans->payload are locked
 * @return checksum in network byte order
 */
uint16_t nip_tcp_checksum( nip_net_if_trans_t *trans )
{
	uint16_t sum    = 0;
// 	uint16_t proto  = NIP_NUM_IP_PROTO_TCP;
	uint16_t tcplen = trans->header_size + trans->payload_size;
	uint16_t word;
	uint16_t *word_ptr; //NOTE don't worry, it's not being used uninitialized
	struct nip_tcp_header *head = nip_mem_obtain_ptr( trans->header  );

	// checkpos describes position in checksum calculation:
	// 0 -> srcaddr1-2
	// 1 -> srcaddr3-4
	// 2 -> dstaddr1-2
	// 3 -> dstaddr3-4
	// 4 -> proto
	// 5 -> tcp length
	// >=6 -> some header pos
	// >=6+headerlen -> payload
	uint8_t  checkpos = 0;

	do
	{
		if      ( checkpos == 0 )
			word_ptr = (uint16_t*)trans->params.ip.src_nw_addr;
		else if ( checkpos == 2 )
			word_ptr = (uint16_t*)trans->params.ip.dst_nw_addr;
/*		else if ( checkpos == 4 )
			word_ptr = &proto;*/
		else if ( checkpos == 6 + trans->header_size )
			word_ptr = nip_mem_obtain_ptr( trans->payload );
		else if ( checkpos == 6 )
			word_ptr = (uint16_t*)head;

		if ( word_ptr == NULL )
			break;

		if ( checkpos == 4 )
			word = NIP_IP_PROTO_TCP;
		else if ( checkpos == 5 )
			word = tcplen;
		else if ( checkpos == 6 + tcplen -1 )
			// uneven number of bytes in payload? -> pad last byte
			word = ((uint8_t*)word_ptr)[0] << 8;
		else
			word = ntohs(*word_ptr);
		sum += word;
		if ( sum < word )
			sum++;

		word_ptr++;
		if ( checkpos >= 6 )
			checkpos++;
		checkpos++;

	} while ( checkpos < tcplen + 6 );

	sum = htons(~sum);

	if ( head != NULL )
		head->checksum = sum;

	nip_mem_release_block( trans->payload );
	nip_mem_release_block( trans->header  );

	return sum;
}


/** Dispatcher function to send TCP segment. The following dispatcher parameters
 * are expected to be configured:
 *   - nip_disp.next.tcp.sockid
 *   - nip_disp.next.tcp.flags
 */
void nip_tcp_disp_send( void )
{
	struct nip_tcb        *tcb;
	nip_net_if_trans_t    *trans;
	struct nip_tcp_header *head;
	uint8_t               rto;
	uint8_t               optlen     = 0; /** make sure it's multiples of 4 */
	uint16_t              new_data_size = 0;
	uint32_t              seq;
	nip_mem_size_t        snd_buf_size;


	tcb = nip_tcp_find_conn( nip_disp.next.tcp.sockid, NULL, 0 );

	if ( tcb == NULL )
		goto quit;

	snd_buf_size = nip_mem_buf_used( tcb->snd_buf );

	// get transmission
	trans = nip_ip_route( tcb->rem_addr.ip, NIP_NET_NO_IF, NULL );
	if ( trans == NULL )
	{
		// out of ressources or no route to host -> try again later
		/// @todo if SYN flag is set and NO_ROUTE_TO_HOST error occured -> cancel
		goto release_connection;
	}

	/// @todo Recognize failed connection attempts when ARP fails? Don't do
	/// automatic reset of transmission status. You have to implement some check
	/// function then, to reset transmission status later on.
	trans->flags.reset_status = 1;

	// build transmission header
	trans->header = nip_mem_alloc(
		NIP_TCP_MIN_HEADER_SIZE,
		NIP_TCP_MIN_HEADER_SIZE,
		0,
		NULL
	);

	if ( trans->header == NIP_MEM_NULL
	  || nip_mem_write( trans->header, NULL, NIP_TCP_MIN_HEADER_SIZE ) != NIP_E_OK )
		goto reset_transmission;

	head = nip_mem_obtain_ptr( trans->header );
	if ( head == NULL )
		goto reset_transmission;

	trans->header_size  = NIP_TCP_MIN_HEADER_SIZE + optlen;
	trans->flags.free_payload = 0;

	// set header flags
	head->flags         = nip_disp.next.tcp.flags;

	if ( tcb->status != NIP_TCP_STAT_SYN_SENT )
	{
		head->flags     |= NIP_TCP_ACK_FLAG;
	}

		// synchronized connections are waiting for ACK.
	if ( tcb->status == NIP_TCP_STAT_SYN_RECEIVED
	  || tcb->status == NIP_TCP_STAT_SYN_SENT
		)
	{
		head->flags  |= NIP_TCP_SYN_FLAG;
	}
	else
	// All data transmitted? Send FIN
	if ( snd_buf_size == 0
	  && ( tcb->status == NIP_TCP_STAT_FIN_WAIT1
	    || tcb->status == NIP_TCP_STAT_LAST_ACK
	    || tcb->status == NIP_TCP_STAT_CLOSING
	    || tcb->status == NIP_TCP_STAT_LAST_ACK )
		)
	{
		head->flags  |= NIP_TCP_FIN_FLAG;
	}
	else
	{
		/// @todo implement check, that TCP segments don't become too large
		/// --> mind tcb->snd_wnd and default-segment-size
		// calculate how much new (not retransmitted) data will be sent
		new_data_size = snd_buf_size - tcb->snd_nxt + tcb->snd_una;
	}

	/// @todo check if it's ok to set seq to snd_nxt for the default case?! If
	/// not: what other default value would seq get?
	seq           = tcb->snd_nxt;

	// send data?
	if ( (tcb->flags & NIP_TCP_SOCK_FLG_SEND) == 0 )
	{
		trans->payload_size = 0;
		trans->payload      = NIP_MEM_NULL;
	}
	else
	{
		if ( head->flags & NIP_TCP_PSH_FLAG )
			tcb->push_ptr = tcb->snd_nxt;

		trans->payload      = tcb->snd_buf;
		// this would be the else-case for the following retransmit condition. To
		// save program memory we omit the else and live with little more
		// processing time.
		trans->payload_off  = snd_buf_size - new_data_size;
		trans->payload_size = new_data_size;

		// payload depends on retransmission
		if ( tcb->rto <= nip_tickcount() )
		{
			trans->payload_off  = 0;
			trans->payload_size = snd_buf_size;
			seq                 = tcb->snd_una;
			if ( tcb->snd_una < tcb->push_ptr )
				head->flags      = nip_disp.next.tcp.flags | NIP_TCP_PSH_FLAG;
			tcb->rt_count++;

			if ( tcb->push_ptr > seq )
				head->flags |= NIP_TCP_PSH_FLAG;
		}
	}

	// Set retransmission timeout
	if ( seq == tcb->snd_una )
	{
		rto = tcb->srtt * NIP_TCP_RTO_BETA;
		if ( rto < NIP_TCP_RTO_LBOUND ) rto = NIP_TCP_RTO_LBOUND;
		if ( rto > NIP_TCP_RTO_UBOUND ) rto = NIP_TCP_RTO_UBOUND;
		tcb->rto = nip_tickcount() + rto;
		tcb->rt_una = seq + trans->payload_size;
		tcb->rt_count = 0;
	}

	// FIN and SYN are special messages, containing no data but increase snd_nxt
	// nevertheless for following messages. In this implementation snd_nxt has
	// been increased prior to calling tcp_disp_snd, so it won't interfere with
	// retransmissions. While waiting for SYN or FIN to be acked we'll still
	// send with the old sequence number, though.
	if ( head->flags & ( NIP_TCP_FIN_FLAG | NIP_TCP_SYN_FLAG ) )
		seq           = tcb->snd_nxt-1;

	head->seq           = htonl( seq );
	head->srcport       = NIP_TCP_SOCKATTR(tcb->sockid,addr.port);
	head->dstport       = tcb->rem_addr.port;
	head->ack           = htonl( tcb->rcv_nxt );
	head->dataoffset    = (trans->header_size / 4 ) << 4;
	/// @todo adjust window size??
	head->window        = htons( tcb->rcv_wnd );
	/// @todo implement urgent pointer

	// calculate checksum
	head->checksum      = 0;
	nip_mem_release_block( trans->header );

	nip_tcp_checksum( trans );

	// increment snd_nxt-pointer
	tcb->snd_nxt    += new_data_size;

	// IP Send
	trans->params.ip.protocol = NIP_IP_PROTO_TCP;
	trans->status   = NIP_NET_IF_RESOLV_ADDR;
	nip_disp.next.common.trans = trans;
	NIP_CURR_CMD    = & nip_ip_disp_send;

	goto release_connection;

reset_transmission:
	trans->payload = NIP_MEM_NULL;
	nip_mem_release_block( trans->header );
	nip_reset_trans( trans );

release_connection:
	nip_mem_release_block( nip_tcp_conns );

quit:
	return;
}


/** Dispatcher function to receive TCP segment.
 * Requires nip_disp.next.common.trans to point to valid transmission
 */
void nip_tcp_disp_receive( void )
{
	nip_net_if_trans_t    *trans = nip_disp.next.common.trans;
// 	nip_time_t            tc     = nip_tickcount();
	uint8_t               flags;
	uint16_t              wnd;
	uint32_t              seq;
	uint32_t              ack;
	uint32_t              nxt;   // next seq after current segment
	uint32_t              eownd; // end of receive window. NOTE Don't worry, it's not being used uninitalized
	nip_mem_size_t        seglen;
	nip_mem_size_t        hlen;  // tcp header length
	nip_mem_size_t        payoff;// payload offset in payload buffer
	nip_mem_size_t        rcvbuf_size;
	nip_mem_handle_t      rcvbuf;
	struct nip_tcp_header *head;
	struct nip_tcb        *tcb = NULL;
	uint8_t               tmp;
	nip_mem_handle_t      hbuf;  // header buffer
#if NIP_TCP_MAX_SOCKETS > 1
	nip_tcp_sock_id_t sockid;
#endif

	// sanity check of payload
	if ( trans->payload_size < NIP_TCP_MIN_HEADER_SIZE )
		goto discard;

	if  ( trans->header != NIP_MEM_NULL )
		hbuf = trans->header;
	else
		hbuf = trans->payload;
	head = nip_mem_obtain_ptr( hbuf );

	if ( head == NULL )
		goto discard;

	// check packet
	hlen   = (head->dataoffset>>4) * 4;
	seglen = trans->payload_size;
	if ( hbuf == trans->payload )
	{
		if ( hlen > trans->payload_size )
		{
			nip_mem_release_block( hbuf );
			goto discard;
		}
		seglen -= hlen;
	}
	///\todo check checksum

	// cache some often needed header variables
	trans->params.tcp.src_port = head->srcport;
	trans->params.tcp.dst_port = head->dstport;
	flags   = head->flags;
	wnd     = ntohs( head->window );
	seq     = ntohl( head->seq );
	ack     = ntohl( head->ack );
	nxt     = seq + seglen;


	// Release and discard header. We saved all we need. Make sure to not read
	// from payload buffer, that shall not be freed
	nip_mem_release_block( hbuf );
	payoff = 0;
	if ( hbuf != trans->payload || trans->flags.free_payload == 1 )
	{
		nip_mem_read( hbuf, NULL, hlen );
	}
	else if ( hbuf == trans->payload )
	{
		payoff = hlen;
	}
	trans->payload_size = payoff + seglen;


	// check if we have socket on destination port
#if NIP_TCP_MAX_SOCKETS > 1
	sockid = 0;
	do
	{
#endif
		if ( NIP_TCP_SOCKATTR(sockid,addr).port == trans->params.tcp.dst_port
		  && ( nip_memcmp( NIP_TCP_SOCKATTR(sockid,addr).ip, trans->params.ip.dst_nw_addr, 4 ) == 0
		    || nip_memcmp( NIP_TCP_SOCKATTR(sockid,addr).ip, nip_ip_null, 4 ) == 0 )
		   )
			goto socketfound;
#if NIP_TCP_MAX_SOCKETS > 1
	} while ( ++sockid < NIP_TCP_MAX_SOCKETS );
#endif
	// socket not found? Discard SYN attempts, Reset everything else.
	if ( flags & NIP_TCP_SYN_FLAG )
		goto discard;
	else
		goto reset;

	socketfound:

	// Read header from payload buffer

	// check for matching connection on that socket
	tcb = nip_tcp_find_conn( sockid, trans->params.ip.src_nw_addr, trans->params.tcp.src_port );

	// P R O C E S S   R E Q U E S T

	// configure parameters for possible replies
	nip_disp.next.tcp.flags  = 0;
	nip_disp.next.tcp.sockid = NIP_TCP_NO_SOCKET;

	if ( tcb != NULL )
	{
		eownd = tcb->rcv_nxt + tcb->rcv_wnd;

		// check if we received RESET
		if ( flags & NIP_TCP_RST_FLAG && tcb->status != NIP_TCP_STAT_LISTEN )
		{
			// check validity of received RESET
			if ( tcb->status == NIP_TCP_STAT_SYN_SENT )
			{
				if ( tcb->snd_nxt != ack )
					goto release_conns;
			}
			else if ( seq < tcb->rcv_nxt || seq > eownd )
				goto release_conns;

			// valid? -> reset connection
			tcb->status = NIP_TCP_STAT_CLOSED;
			nip_mem_release_block( nip_tcp_conns );
			goto release_conns;
		}
	}

	// Update connection
	if ( flags & NIP_TCP_ACK_FLAG )
	{
		// Received segment is ACKing a non-existent connetion?
		// -> Tell remote host to reset his connection.
		if ( tcb == NULL || tcb->status == NIP_TCP_STAT_LISTEN )
			goto reset;

		// received acceptable ACK?
		if ( tcb->snd_una < ack && ack <= tcb->snd_nxt )
		{
			// remove acked data from transmission buffer
			if ( tcb->flags & NIP_TCP_SOCK_FLG_SEND )
			{
				nip_mem_read( tcb->snd_buf, NULL, ack - tcb->snd_una );

				// Send buffer emptied? -> check FIN states
				if ( nip_mem_buf_used( tcb->snd_buf ) == 0
				 &&  tcb->snd_una != tcb->snd_nxt
				 &&  ( tcb->status == NIP_TCP_STAT_FIN_WAIT1
				    || tcb->status == NIP_TCP_STAT_LAST_ACK )
					)
				{
					// increment snd_nxt and send FIN
					tcb->flags   &= ~NIP_TCP_SOCK_FLG_SEND;
					tcb->snd_nxt++;
					NIP_CURR_CMD             = &nip_tcp_disp_send;
				}
			}

			tcb->snd_una = ack;

			// check retransmission queue
			if ( ack >= tcb->rt_una )
			{
				/// @todo delete top of queue. Current implementation has no queue
				/// and therefor assumes next item to be snd_nxt, which is not the
				/// case if there have been multiple transmissions since. Also there
				/// is nothing known about the original transmission time of the
				/// next segment, to assume the current time is wrong. We're doing
				/// that to save program memory, as TCP will also work this way. It
				/// might just perform slower, as retransmissions of subsequent
				/// segments may be delayed.
				tcb->rt_una = tcb->snd_nxt;
				tcb->rto    = nip_tickcount() + NIP_TCP_RTO_LBOUND;

				/// @todo perform roundtriptime calculation here
			}

		}

		// if data has been received, update remote peer with current status
		if ( seglen > 0 )
			NIP_CURR_CMD             = &nip_tcp_disp_send;

		// received valid data?
		if ( (tcb->rcv_wnd == 0 && seglen == 0 && tcb->rcv_nxt == seq)
		  || ( tcb->rcv_wnd >  0
		    && ( (seq <= tcb->rcv_nxt )
		      && (tcb->rcv_nxt <= nxt && nxt <= eownd )
		       )
			  )
			)
		{
			// make sure connection is not synchronizing or closing
			if ( tcb->flags & NIP_TCP_SOCK_FLG_RECV )
			{
				// Try to add data to socket. This may fail if receive-buffer
				// is too small and can also not be increased due to locked buffers.
				// To minimize that risk, we release (unlock) the connection buffer
				// so it won't interrupt that operation.
				rcvbuf      = tcb->rcv_buf;
				rcvbuf_size = nip_mem_buf_used( tcb->rcv_buf );
				// don't read bytes, that have already been received
				payoff += tcb->rcv_nxt - seq;

				if ( (tcb->flags & NIP_TCP_SOCK_FLG_SHUT_RD) == 0)
				{
					nip_mem_release_block( nip_tcp_conns );

					if ( trans->flags.free_payload == 1 )
					{
						// move data from transmission payload to socket buffer.
						// first dump data we have already received
						nip_mem_read( trans->payload, NULL, payoff );
						seglen = nip_mem_move( rcvbuf, trans->payload, seglen - payoff );
					}
					else
					{
						// determine number of bytes to be copied
						seglen = trans->payload_size - payoff;

						// perform slow copy operation
						do
						{
							if ( 1 != nip_mem_read_at_pos( trans->payload, &tmp, 1, payoff++ )
							|| NIP_E_OK != nip_mem_write( rcvbuf, &tmp, 1 ) )
							{
								break;
							}
						} while ( payoff < trans->payload_size );

						// subtract number of bytes not copied, so we won't ACK those
						seglen -= (trans->payload_size - payoff);
					}

					tcb = nip_tcp_find_conn( sockid, trans->params.ip.src_nw_addr, trans->params.tcp.src_port );
					// since we just had access to the TCB there should be no problem
					// accessing it again. If we do that's a programming error in the
					// stack.
					if ( tcb == NULL )
					{
						// remove the newly read data from the receive buffer, because
						// we're unable to update the TCB
						nip_mem_set_used( rcvbuf, rcvbuf_size );
						goto discard;
					}
				}

				tcb->rcv_nxt     += seglen;

				/// @TODO update receive window?

			}
		}
	}

	// synchronize
	if ( flags & NIP_TCP_SYN_FLAG )
	{
		// No connection found? attempt to create new connection
		if ( tcb == NULL
		  && (NIP_TCP_SOCKATTR(sockid,flags) & NIP_TCP_SOCK_FLG_LISTENING)
		  &&  nip_tcp_unaccepted_cnt < NIP_TCP_MAX_UNACCEPTED )
		{
			tcb = nip_tcp_create_conn(
				sockid,
				trans->params.ip.src_nw_addr,
				trans->params.tcp.src_port,
				NIP_TCP_STAT_LISTEN
			);
			tcb->flags = NIP_TCP_SOCKATTR(sockid,flags);
			nip_tcp_unaccepted[nip_tcp_unaccepted_cnt] = tcb->connid;
			nip_tcp_unaccepted_cnt++;
		}
		// we should now have a pointer to a valid connection control block for
		// the incoming packet. If we don't, we ran out of ressources and discard.
		if ( tcb == NULL )
			goto discard;

		// update connection status for LISTEN or SYN_SENT connections
		if ( tcb->status == NIP_TCP_STAT_LISTEN
		  || tcb->status == NIP_TCP_STAT_SYN_SENT )
		{
			tcb->irs     = seq;
			tcb->rcv_nxt = tcb->irs + 1;
			// wl1 and wl2 will help to prevent using old segments to update
			// the window (snd_wnd)
			tcb->snd_wl1 = seq;
			tcb->snd_wl2 = ack;
			tcb->snd_wnd = wnd;

			tcb->status  = NIP_TCP_STAT_SYN_RECEIVED;
			///@todo check options
			///@todo tcb->snd_maxsegsize = ?
		}

		// Reply to any SYN with current system state, even if SYN was
		// unexpected.
		NIP_CURR_CMD             = &nip_tcp_disp_send;
	}

	// received last data?
	if  ( flags & NIP_TCP_FIN_FLAG )
	{
		if ( tcb == NULL )
			goto reset;

		// check that we've received everything we should
		if ( tcb->rcv_nxt == seq )
		{

			// setup connection to not receive any more data
			tcb->flags &= ~NIP_TCP_SOCK_FLG_RECV;

			if ( tcb->status == NIP_TCP_STAT_ESTABLISHED )
				tcb->status = NIP_TCP_STAT_CLOSE_WAIT;

			if ( tcb->status == NIP_TCP_STAT_FIN_WAIT1 )
				tcb->status = NIP_TCP_STAT_CLOSING;

			if ( tcb->status == NIP_TCP_STAT_FIN_WAIT2 )
			{
				tcb->timeout = nip_tickcount() + NIP_TCP_MSL;
				tcb->status  = NIP_TCP_STAT_TIMEWAIT;
			}

			tcb->rcv_nxt++;
		}

		// acknowledge FIN
		NIP_CURR_CMD = &nip_tcp_disp_send;
	}


	// Perform state transitions after last data has been acked
	if ( tcb->snd_una == tcb->snd_nxt )
	{
		if ( tcb->status == NIP_TCP_STAT_SYN_RECEIVED )
		{
			tcb->status = NIP_TCP_STAT_ESTABLISHED;
			tcb->flags |= NIP_TCP_SOCK_FLG_SEND | NIP_TCP_SOCK_FLG_RECV;
		}

		if ( tcb->status == NIP_TCP_STAT_CLOSING )
		{
			tcb->timeout = nip_tickcount() + NIP_TCP_MSL;
			tcb->status  = NIP_TCP_STAT_TIMEWAIT;
		}

		if ( tcb->status == NIP_TCP_STAT_FIN_WAIT1 )
			tcb->status = NIP_TCP_STAT_FIN_WAIT2;

		if ( tcb->status == NIP_TCP_STAT_LAST_ACK )
			tcb->status = NIP_TCP_STAT_CLOSED;
	}

	if ( tcb != NULL )
		nip_disp.next.tcp.sockid = tcb->connid;


	release_conns:
		nip_mem_release_block( nip_tcp_conns );


	// D I S C A R D
	discard:
		trans->status = NIP_NET_IF_DONE;
		nip_reset_trans( trans );
		return;


	// S E N D   R E S E T
	reset:

		if ( tcb != NULL )
			nip_mem_release_block( nip_tcp_conns );

		// send RST segment based on current transmission

		trans->payload_size = 0;

		// get pointer to header
		if ( trans->header == NIP_MEM_NULL )
		{
			// build transmission header
			trans->header = nip_mem_alloc(
				NIP_TCP_MIN_HEADER_SIZE,
				NIP_TCP_MIN_HEADER_SIZE,
				0,
				NULL
			);

		}
		else
		{
			// use given payload buffer for header, as we're not sending any payload
			nip_mem_set_used( trans->header, 0 );
		}

		// Fill header buffer with Zeros
		if ( trans->header == NIP_MEM_NULL
		|| nip_mem_write( trans->header, NULL, NIP_TCP_MIN_HEADER_SIZE ) != NIP_E_OK )
		{
			goto discard;
		}

		head = nip_mem_obtain_ptr( trans->header );
		if ( head == NULL )
		{
			goto discard;
		}

		trans->header_size  = NIP_TCP_MIN_HEADER_SIZE;

		// write header
		// swap ports from original transmission
		nip_memcpy( trans->params.tcp.dst_nw_addr, trans->params.tcp.src_nw_addr, 4 );
		head->srcport    = trans->params.tcp.dst_port;
		head->dstport    = trans->params.tcp.src_port;
		head->dataoffset = NIP_TCP_MIN_DATA_OFFSET;
		head->flags      = NIP_TCP_RST_FLAG;
		head->ack        = NIP_TCP_RESET_ACK;
		if ( flags & NIP_TCP_ACK_FLAG )
			head->seq        = htonl(ack);

		nip_mem_release_block( trans->header );

		trans = nip_ip_route( trans->params.tcp.dst_nw_addr, NIP_NET_NO_IF, trans );
		if ( trans == NULL )
		{
			// no route to host? whatever... it's just a reset we're losing.
			goto discard;
		}
		nip_tcp_checksum( trans );

		// IP Send
		trans->params.ip.protocol = NIP_IP_PROTO_TCP;
		trans->status   = NIP_NET_IF_RESOLV_ADDR;
		NIP_CURR_CMD    = & nip_ip_disp_send;

		return;
}

/** dispatcher function to remove closed connection
 */
void nip_tcp_disp_close( void )
{
	nip_tcp_del_conn( nip_disp.next.tcp.sockid );
}



/** Dispatcher function to check TCP connection states/retransmissions/..
 */
void nip_tcp_disp_check( void )
{
	struct nip_tcb     *tcb;
	nip_tcp_sock_cnt_t cnt = 0;
	nip_mem_size_t     used;

	if ( nip_tcp_conn_cnt > 0 )
	{
		// get pointer to tcp array
		tcb = nip_mem_obtain_ptr( nip_tcp_conns );
		if ( tcb != NULL )
		{
			// check all tcp connections
			do
			{
				// if we need to move on before all all connections have been
				// checked, notify dispatcher to come back.
				if ( NIP_CURR_CMD != NULL )
				{
					nip_disp_notify( NIP_DISP_CHECK_TCP );
					break;
				}

				nip_disp.next.tcp.flags  = 0;
				nip_disp.next.tcp.sockid = tcb->connid;

				used = nip_mem_buf_used( tcb->snd_buf );

				// check for closed connections
				if ( tcb->timeout >= nip_tickcount()
				  && tcb->status == NIP_TCP_STAT_TIMEWAIT )
				{
					tcb->status = NIP_TCP_STAT_CLOSED;
				}

				if ( tcb->status == NIP_TCP_STAT_CLOSED
				  && tcb->flags & NIP_TCP_SOCK_FLG_AUTOCLOSE )
				{
					NIP_CURR_CMD = &nip_tcp_disp_close;
				}
				else
				// check for data or FIN that needs to be transmitted
				if ( (used  > 0 && used > tcb->snd_nxt - tcb->snd_una)
				  || (used == 0
				    && ( tcb->status == NIP_TCP_STAT_FIN_WAIT1
				      || tcb->status == NIP_TCP_STAT_LAST_ACK )
				     )
				  || ( tcb->snd_nxt > tcb->snd_una && tcb->rto <= nip_tickcount() )
					)
					///@todo insert congestion control
				{
					NIP_CURR_CMD = &nip_tcp_disp_send;
				}
				tcb++;
			} while ( ++cnt < nip_tcp_conn_cnt );

			// done? -> free connection pointer
			nip_mem_release_block( nip_tcp_conns );
		}
	}

}

#endif /* NIP_TCP_ENABLE */
