/*##############################################################################

nIP - nano IP stack

File        : udp.c

Description : UDP

Copyright notice:

Copyright (C) 2005 -
Andreas Dittrich, dittrich@informatik.hu-berlin.de
Jon Kowal, kowal@informatik.hu-berlin.de

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#############################################################################*/

#include "nip_init.h"
#include "dispatcher.h"
#include "inet.h"
#include "os_core.h"
#include "nip_error.h"
#include "net/net_if.h"
#include "net/udp.h"


#if (NIP_UDP_ENABLE == 1 )

#if (NIP_UDP_USE_FIXED_SOCKETS == 1)
/** Fixed array of UDP sockets */
static struct nip_udp_sock   udp_sockets[ NIP_UDP_FIXED_SOCKETS ];
#else

/** Handle of reserved memory area for UDP Sockets */
static nip_udp_sock_cnt_t udp_socket_cnt;
static nip_mem_handle_t   udp_sockets;
#endif // NIP_UDP_USE_FIXED_SOCKETS == 1

/** Fixed array of UDP queue */
static struct nip_udp_buffer udp_rx_queue[ NIP_UDP_PACKET_QUEUE_SIZE ];

/** Register UDP socket with given address and flags.
 * @param addr structure to be filled with address and port. If all address
 *             bytes are set to Zero, the socket will listen on all interfaces.
 * @param flags possible flags are:
 *               - NIP_UDP_SOCK_FLG_NON_BLOCK
 *               - NIP_UDP_SOCK_FLG_LISTENING
 * @return NIP_UDP_NO_SOCKET on error. Global nip_error variable will be set to
 * NIP_E_OUT_OF_RESSOURCES in that case.
 */
nip_udp_sock_id_t nip_udp_socket( struct nip_udp_sock_addr *addr, uint8_t flags )
{
	nip_udp_sock_id_t   id    = 0;
	struct nip_udp_sock *sock = NULL;
	///@todo reserve socket in global list
#if (NIP_UDP_USE_FIXED_SOCKETS==1)
	do
	{
		if ( (udp_sockets[ id ].flags & NIP_UDP_SOCK_FLG_USED) == 0 )
		{
			sock = & udp_sockets[ id ];
			sock->sockid = id;
			sock->flags  = flags | NIP_UDP_SOCK_FLG_USED;
			nip_memcpy( &sock->addr, addr, sizeof( struct nip_udp_sock_addr ) );
			sock->rx_packet = NIP_UDP_EOQ;
			sock->tx_packet = NIP_MEM_NULL;
			return id;
		}
	} while ( ++id < NIP_UDP_FIXED_SOCKETS );

#else
	///@todo implement not fixed sockets
#endif // (NIP_UDP_USE_FIXED_SOCKETS==1)


	nip_error = NIP_E_OUT_OF_RESSOURCES;
	return NIP_UDP_NO_SOCKET;

}

/** Unregister previous socket registration */
void nip_udp_close( nip_udp_sock_id_t id )
{
	struct nip_udp_buffer *qbuf = udp_rx_queue;

	// empty receive queue for that socket
	do
	{
		if ( qbuf->sockid == id )
		{
			nip_mem_free( qbuf->buf );
			qbuf->buf = NIP_MEM_NULL;
		}
	} while ( ++qbuf != &udp_rx_queue[NIP_UDP_EOQ] );

#if (NIP_UDP_USE_FIXED_SOCKETS==1)
	if ( id >= NIP_UDP_FIXED_SOCKETS ) return;

	nip_mem_free( udp_sockets[id].tx_packet );
	udp_sockets[ id ].flags = 0;
#else
	///@todo implement not fixed sockets
#endif // (NIP_UDP_USE_FIXED_SOCKETS==1)
}

/** Return pointer to socket structure belonging to given sockid or NULL if
 * socket was not found.
 */
struct nip_udp_sock *nip_udp_sock_ptr ( nip_udp_sock_id_t sockid )
{
#if (NIP_UDP_USE_FIXED_SOCKETS==1)
	if ( sockid < NIP_UDP_FIXED_SOCKETS && udp_sockets[ sockid ].flags & NIP_UDP_SOCK_FLG_USED )
		return &udp_sockets[ sockid ];
	else
	{
		nip_error = NIP_E_INVALID_SOCK;
		return NULL;
	}
#else
	///@todo implement not fixed sockets
#endif // (NIP_UDP_USE_FIXED_SOCKETS==1)
}

/** Return handle to current Receive or Transmit buffer of UDP socket
 */
nip_mem_handle_t nip_udp_buf_handle( nip_udp_sock_id_t sockid, nip_udp_buf_type_t buftype )
{
	struct nip_udp_sock *sock = nip_udp_sock_ptr( sockid );
	nip_mem_handle_t    res   = NIP_MEM_NULL;

	if ( sock != NULL )
	{
		res = sock->tx_packet;

		if ( buftype == NIP_UDP_RX_BUF )
		{
				if ( sock->rx_packet != NIP_UDP_EOQ )
					res       = udp_rx_queue[ sock->rx_packet ].buf;
				else
				{
					nip_error = NIP_E_INVALID_BLOCK;
				}
		}
	}
	else
		nip_error = NIP_E_INVALID_SOCK;

	return res;
}

/** Release previously granted direct access.
 *
 * This function has to be used in combination with nip_udp_data_ptr() as all all
 * pointers obtained by the latter have to be released again.
 *
 * @warning Do not use the given direct-access pointer after calling this
 * function.
 */
void nip_udp_ptr_release( nip_udp_sock_id_t sockid, nip_udp_buf_type_t buftype )
{
	nip_mem_handle_t    bufhandle = nip_udp_buf_handle( sockid, buftype );

	if ( bufhandle != NIP_MEM_NULL )
		nip_mem_release_block( bufhandle );
}

/** Return mem-ptr to buffer data. The call failed, if res->id is NIP_MEM_NULL.
 */
void nip_udp_data_mem_ptr( nip_udp_sock_id_t sockid, nip_udp_buf_type_t buftype, nip_mem_ptr_t *res )
{
	res->id  = nip_udp_buf_handle( sockid, buftype );
	if ( res->id != NIP_MEM_NULL )
		res->ptr = (void*) sizeof( struct nip_udp_header );
	else
		res->ptr = NULL;
}

/** Give direct access (pointer) to a socket buffer.
 *
 * Normally the buffer space is administered by the UDP implementation and the
 * application only has write access through the nip_udp_write() function. For
 * optimization reasons it may be better for the application to have direct
 * access, though.
 *
 * The direct TX buffer access is intended to be used in combination with the
 * nip_udp_write() function, which still has to be used to allocate enough
 * memory. You may call nip_udp_write() without supplying a buffer pointer to
 * allocate uninitialized memory.
 *
 * If the requested buffer size could not be allocated NULL will returned and
 * nip_error be set to NIP_E_OUT_OF_MEMORY.
 *
 * @warning The buffer will be locked until nip_udp_ptr_release() has been called
 * which will affect memory management and packet transmission. You have to call
 * nip_udp_ptr_release() before transmitting the packet.
 *
 * @warning Make sure you know how much memory has been allocated previously
 * (using the nip_udp_write() function). You MUST NOT write beyond that mark or
 * buffer overflows in the memory management system will occur.
 *
 * @param sockid id of socket to return pointer at buffer for
 *
 * @return pointer to buffer or NULL. In the latter case nip_error
 * NIP_E_INVALID_SOCK for an invalid sockid or NIP_E_INVALID_BLOCK if the
 * buffer had not been initialized, or NIP_E_BLOCK_LOCKED if ressource was
 * locked.
 */
void *nip_udp_data_ptr( nip_udp_sock_id_t sockid, nip_udp_buf_type_t buftype )
{
	nip_mem_ptr_t  ptr;

	nip_udp_data_mem_ptr( sockid, buftype, &ptr );

	return nip_mem_ptr( &ptr );
}

/** Initialize send buffer for new packet. Use this function in combination with
 * nip_udp_write() or nip_udp_data_ptr() and nip_udp_transmit().
 *
 * @param sockid    ID of socket to initialize send buffer for.
 * @param size      Size to be reserved (may be 0 for dynamic allocation)
 *
 * The socket buffer will be initialized without data, use nip_udp_write() to
 * fill the buffer. You should provide the buffer size, if known in advance, to
 * ensure there is enough memory available to buffer your packet. If you do
 * not provide the buffer size it will be automatically increased whenever
 * nip_udp_write() is being called, which may then fail if not enough memory is
 * available.
 *
 * Call nip_udp_transmit() to send the buffer. nip_udp_data_ptr() will hand you a
 * pointer to the buffer. See the description of nip_udp_data_ptr() for further
 * instructions and warnings!
 *
 * @warning Any previous buffer that has not been scheduled for transmission
 * will be overwritten.
 *
 * @return NIP_E_OK, NIP_E_OUT_OF_MEMORY, of NIP_E_INVALID_SOCK
 */
nip_error_t nip_udp_init_send( nip_udp_sock_id_t sockid, nip_mem_size_t size )
{
	struct nip_udp_sock   *sock = nip_udp_sock_ptr( sockid );

	if ( sock == NULL )
	{
		return NIP_E_INVALID_SOCK;
	}

	// add UDP header to packet size
	size += sizeof ( struct nip_udp_header );
	sock->tx_packet = nip_mem_alloc( size, size, NIP_MEM_FLG_DELREAD, NULL );

	if ( sock->tx_packet == NIP_MEM_NULL )
	{
		return NIP_E_OUT_OF_MEMORY;
	}

	// add UDP header space to buffer without initializing any values (that will
	// be done in nip_udp_transmit().
	nip_mem_write( sock->tx_packet, NULL, sizeof( struct nip_udp_header ) );

	return NIP_E_OK;
}

/** Cancel transmission. The send buffer will be cleared and removed. For the
 * next packet to be sent, nip_udp_init_send() will have to be called.
 *
 * This function will usually be called by the application after nip_udp_write()
 * failed before the packet was complete, which normally is a caused by
 * insufficient memory. If the application does not want to wait in place for
 * enough memory to be available and has no means to safe the send-status it
 * will cancel the transmission and retry later.
 */
void nip_udp_cancel_send( nip_udp_sock_id_t sockid )
{
	struct nip_udp_sock   *sock = nip_udp_sock_ptr( sockid );

	nip_mem_free( sock->tx_packet );
	sock->tx_packet = NIP_MEM_NULL;
}

/** Write bytes to socket. Transmission has to be trigged by nip_udp_transmit().
 */
nip_error_t nip_udp_write( nip_udp_sock_id_t sockid, uint8_t *buf, nip_mem_size_t size)
{
	struct nip_udp_sock *sock = nip_udp_sock_ptr( sockid );

	//copy bytes to transmission buffer
	return nip_mem_write( sock->tx_packet, buf, size );
}

/** Transmit the packet from the current send buffer to the given destination.
 *
 * @warning Never perform transmit calls from within the dispatcher on blocking
 * sockets!!! That will most likely lead to deadlock as nip_udp_transmit() will
 * prevent the dispatcher from finishing the transmission until it has finished
 * the transmission, which would happen in the not so near future near infinity.
 * ;)
 *
 * @param sockid     Socket ID
 * @param *dest_addr pointer to destination address structure
 * @param net_if     ID of network interface to transmit message on. If you do
 *                   not wish to specify the interface, put NIP_NET_NO_IF.
 *
 * @return
 *   - NIP_E_OK               packet has been added to outgoing packet queue
 *   - NIP_E_DEV_BUSY         outgoing queue of interface is full, try again
 *   - NIP_E_NO_ROUTE_TO_HOST no interface will deliver to destination address
 */
nip_error_t nip_udp_transmit( nip_udp_sock_id_t sockid,struct nip_udp_sock_addr *dest_addr, nip_net_if_id_t if_id )
{
 	struct nip_udp_sock   *sock = nip_udp_sock_ptr( sockid );
	struct nip_udp_header *head;
	nip_net_if_trans_t    *trans;

	if ( sock == NULL )
		return NIP_E_INVALID_SOCK;

	// Call IP route to get transmission pointer
	trans = nip_ip_route( dest_addr->ip, if_id, NULL );
	if ( trans == NULL )
	{
		return nip_error;
	}

	// add UDP packet to transmission and configure transmission to be freed
	// after sending.
	trans->payload = sock->tx_packet;
	trans->payload_size = nip_mem_buf_used( sock->tx_packet );
	trans->flags.free_payload = 1;
	trans->flags.reset_status = 1;
	trans->params.ip.protocol = NIP_IP_PROTO_UDP;

	// get pointer to packet buffer and fill UDP header
	head = nip_mem_obtain_ptr( sock->tx_packet );
	head->src_port = sock->addr.port;
	head->dst_port = dest_addr->port;
	head->length   = htons( trans->payload_size );

	// add pseudo header to checksum
	///@todo add checksum
// 	head->checksum = NIP_NUM_IP_PROTO_UDP +
	head->checksum = 0;

	// release pointer to UDP header
	nip_mem_release_block( sock->tx_packet );


	// reset socket transmission status
	sock->tx_packet = NIP_MEM_NULL;

	// Notify dispatcher to continue IP transmission
	trans->status = NIP_NET_IF_RESOLV_ADDR;
	nip_disp_notify( NIP_DISP_CHECK_TRANS );
	nip_dispatcher();

	// all went well
	return NIP_E_OK;
}


/** Attempt to receive new packet from UDP socket. The payload will be written
 * to the supplied buffer pointer. If the datagram is to large to fit in the
 * supplied buffer, it will be truncated at the buffer boundary, but not
 * discarded. Subsequent calls to nip_udp_recvfrom() may be performed to read the
 * remaining data.
 *
 * @param sockid     id of socket to read from
 * @param *buf       pointer to buffer to write payload to
 * @param size       size of supplied buffer
 * @param *from_addr pointer to structure where sender address information will
 *                   be copied to. This pointer may be NULL if you're not
 *                   interested in address information.
 *
 * @return total size of available buffer. If the return value is smaller than
 * the provided buffer it indicates how much of the buffer has been filled, if
 * it is larger it indicates that the packet has been truncated. If 0 is returned
 * the nip_error variable will be set to one of the following values:
 *  - NIP_E_AGAIN        no data available on non-blocking socket
 *  - NIP_E_INVALID_SOCK invalid socket or socket closed while waiting for data
 */
nip_mem_size_t nip_udp_recvfrom( nip_udp_sock_id_t sockid, uint8_t *buf, nip_mem_size_t size, struct nip_udp_sock_addr *from_addr )
{
 	struct nip_udp_sock   *sock;
	nip_mem_size_t        res;
	struct nip_udp_buffer *qbuf;

	do
	{
		// as the socket may be shut down while the loop is running we'll have to
		// check on it every time.
		sock = nip_udp_sock_ptr( sockid );
		if ( sock == NULL )
		{
			nip_error = NIP_E_INVALID_SOCK;
			return 0;
		}

		// don't block for non-blocking sockets
		if ( sock->rx_packet == NIP_UDP_EOQ && (sock->flags & NIP_UDP_SOCK_FLG_NON_BLOCK) )
		{
			nip_error = NIP_E_AGAIN;
			return 0;
		}

	} while ( sock->rx_packet == NIP_UDP_EOQ );

	qbuf = &udp_rx_queue[ sock->rx_packet ];

	// copy remote address data
	if ( from_addr != NULL )
		nip_memcpy( from_addr, &qbuf->addr, sizeof( struct nip_udp_sock_addr ) );

	//copy bytes from packet to buffer
	size = nip_mem_read( qbuf->buf, buf, size );

	res = qbuf->size;
	qbuf->size -= size;

	// buffer has been emptied? remove item from queue
	if ( res <= size )
	{
		nip_udp_cancel_recv( sockid );
	}

	return res;

}

/** Forwards current incoming buffer to outgoing queue. It will still have to
 * be transmitted using nip_udp_transmit().
 */
void nip_udp_forward( nip_udp_sock_id_t sockid )
{
	struct nip_udp_sock   *sock = nip_udp_sock_ptr( sockid );
	struct nip_udp_buffer *qbuf;

	if ( sock!= NULL && sock->rx_packet != NIP_UDP_EOQ )
	{
		qbuf = &udp_rx_queue[ sock->rx_packet ];
		// clear current transmission
		nip_mem_free( sock->tx_packet );
		sock->tx_packet = qbuf->buf;
		qbuf->buf = NIP_MEM_NULL;
		sock->rx_packet = qbuf->next;
	}
}

/** Cancel current packet reception. */
void nip_udp_cancel_recv( nip_udp_sock_id_t sockid )
{
	struct nip_udp_sock   *sock = nip_udp_sock_ptr( sockid );
	struct nip_udp_buffer *qbuf;

	if ( sock->rx_packet != NIP_UDP_EOQ )
	{
		qbuf = &udp_rx_queue[ sock->rx_packet ];
		// forward to next packet in queue
		sock->rx_packet = qbuf->next;
		// clear old queue item
		nip_mem_free( qbuf->buf );
		qbuf->buf = NIP_MEM_NULL;
	}
}

/** Dispatcher function to be called upon packet reception */
void nip_udp_disp_receive( void )
{
	struct nip_udp_header *head;
	struct nip_udp_sock   *sock;
	nip_net_if_trans_t    *trans;
	nip_udp_sock_id_t     sockid;
	uint8_t               check_cnt;
	nip_udp_queue_id_t    sock_qid;
	nip_udp_queue_id_t    ins_qid;
	struct nip_udp_buffer *qbuf;
	uint16_t              h_size;

	trans = nip_disp.next.common.trans;

	if ( trans->payload_size < sizeof( struct nip_udp_header) )
		goto discard;

	head = nip_mem_obtain_ptr( trans->payload );

	if ( head == NULL )
	{
		// something's wrong with the payload -> discard
		goto discard;
	}

	// lookup socket
#if (NIP_UDP_USE_FIXED_SOCKETS==1)
	sockid = 0;
		#if (NIP_UDP_FIXED_SOCKETS > 1 )
	do
	{
	#endif
		check_cnt = 0;
		sock = &udp_sockets[ sockid ];
		// check socket
		if (((sock->flags & NIP_UDP_SOCK_FLG_USED) > 0)
			&& ( sock->addr.port == head->dst_port )
			&& ((nip_memcmp( sock->addr.ip, trans->params.ip.dst_nw_addr, 4 ) == 0 )
			  ||(sock->addr.ip[0]==0 && sock->addr.ip[1]==0 && sock->addr.ip[2]==0 && sock->addr.ip[3]==0)
			#if NIP_MULTICAST_ENABLE == 1
			  ||(trans->params.ip.dst_nw_addr[0] >= 0xE0)
			#endif
				))
		{
			goto found;
		}
	#if (NIP_UDP_FIXED_SOCKETS > 1 )
	} while ( ++sockid < NIP_UDP_FIXED_SOCKETS );
	#endif
#else
	///@todo implement not fixed sockets
#endif // (NIP_UDP_USE_FIXED_SOCKETS==1)
	goto release;

	found:
	h_size = ntohs(head->length);
	// sanity-check of datagram size
	if ( nip_mem_buf_used( trans->payload ) < h_size )
	{
		goto release;
	}

	// find slot in queue
	ins_qid = 0;
	do
	{
		qbuf = &udp_rx_queue[ ins_qid ];
		if ( qbuf->buf == NIP_MEM_NULL )
			break;
	} while ( ++ins_qid < NIP_UDP_EOQ );

	// queue full? -> discard
	if ( ins_qid == NIP_UDP_EOQ )
		goto release;

	// find end of socket queue
	sock_qid = sock->rx_packet;
	while ( sock_qid != NIP_UDP_EOQ )
	{
		if ( udp_rx_queue[ sock_qid ].next == NIP_UDP_EOQ ) break;
		sock_qid = udp_rx_queue[ sock_qid ].next;
	}

	// set up queue item
	qbuf->buf       = trans->payload;
	qbuf->next      = NIP_UDP_EOQ;
	qbuf->sockid    = sockid;
	qbuf->size      = h_size;
	qbuf->addr.port = head->src_port;
	nip_memcpy( qbuf->addr.ip, trans->params.ip.src_nw_addr, 4 );
	// trans->payload now managed by socket. Make sure it won't be freed
	// with the rest of the transmission
	trans->flags.free_payload = 0;


	// add to queue
	if ( sock->rx_packet == NIP_UDP_EOQ )
	{
		sock->rx_packet = ins_qid;
	}
	else
	{
		udp_rx_queue[ sock_qid ].next = ins_qid;
	}
	// else: queue full? -> Discard datagram
	///@todo implement longer UDP receive Queue

	release:
		nip_mem_release_block( trans->payload );

	discard:
		trans->status = NIP_NET_IF_DONE;
		nip_reset_trans( trans );
		return;
}


#endif // (NIP_UDP_ENABLE == 1 )
